# Authors: Thierry Moudiki
#
# License: BSD 3 Clear Clause

from ..deep import DeepRegressor
from ..mts import MTS


class DeepMTS(MTS):
    """Univariate and multivariate time series (DeepMTS) forecasting with Quasi-Randomized networks (Work in progress /!\)

    Parameters:

        obj: object.
            any object containing a method fit (obj.fit()) and a method predict
            (obj.predict()).

        n_layers: int.
            number of layers in the neural network.

        n_hidden_features: int.
            number of nodes in the hidden layer.

        activation_name: str.
            activation function: 'relu', 'tanh', 'sigmoid', 'prelu' or 'elu'.

        a: float.
            hyperparameter for 'prelu' or 'elu' activation function.

        nodes_sim: str.
            type of simulation for the nodes: 'sobol', 'hammersley', 'halton',
            'uniform'.

        bias: boolean.
            indicates if the hidden layer contains a bias term (True) or not
            (False).

        dropout: float.
            regularization parameter; (random) percentage of nodes dropped out
            of the training.

        direct_link: boolean.
            indicates if the original predictors are included (True) in model's fitting or not (False).

        n_clusters: int.
            number of clusters for 'kmeans' or 'gmm' clustering (could be 0: no clustering).

        cluster_encode: bool.
            defines how the variable containing clusters is treated (default is one-hot)
            if `False`, then labels are used, without one-hot encoding.

        type_clust: str.
            type of clustering method: currently k-means ('kmeans') or Gaussian
            Mixture Model ('gmm').

        type_scaling: a tuple of 3 strings.
            scaling methods for inputs, hidden layer, and clustering respectively
            (and when relevant).
            Currently available: standardization ('std') or MinMax scaling ('minmax').

        lags: int.
            number of lags used for each time series.

        replications: int.
            number of replications (if needed, for predictive simulation). Default is 'None'.

        kernel: str.
            the kernel to use for residuals density estimation (used for predictive simulation). Currently, either 'gaussian' or 'tophat'.

        agg: str.
            either "mean" or "median" for simulation of bootstrap aggregating

        seed: int.
            reproducibility seed for nodes_sim=='uniform' or predictive simulation.

        backend: str.
            "cpu" or "gpu" or "tpu".

        verbose: int.
            0: not printing; 1: printing

        show_progress: bool.
            True: progress bar when fitting each series; False: no progress bar when fitting each series

    Attributes:

        fit_objs_: dict
            objects adjusted to each individual time series

        y_: {array-like}
            DeepMTS responses (most recent observations first)

        X_: {array-like}
            DeepMTS lags

        xreg_: {array-like}
            external regressors

        y_means_: dict
            a dictionary of each series mean values

        preds_: {array-like}
            successive model predictions

        preds_std_: {array-like}
            standard deviation around the predictions

        return_std_: boolean
            return uncertainty or not (set in predict)

        df_: data frame
            the input data frame, in case a data.frame is provided to `fit`

    Examples:

    Example 1:

        ```python
        import nnetsauce as ns
        import numpy as np
        from sklearn import linear_model
        np.random.seed(123)

        M = np.random.rand(10, 3)
        M[:,0] = 10*M[:,0]
        M[:,2] = 25*M[:,2]
        print(M)

        # Adjust Bayesian Ridge
        regr4 = linear_model.BayesianRidge()
        obj_DeepMTS = ns.DeepMTS(regr4, lags = 1, n_hidden_features=5)
        obj_DeepMTS.fit(M)
        print(obj_DeepMTS.predict())

        # with credible intervals
        print(obj_DeepMTS.predict(return_std=True, level=80))

        print(obj_DeepMTS.predict(return_std=True, level=95))
        ```

    Example 2:

        ```python
        import nnetsauce as ns
        import numpy as np
        from sklearn import linear_model

        dataset = {
        'date' : ['2001-01-01', '2002-01-01', '2003-01-01', '2004-01-01', '2005-01-01'],
        'series1' : [34, 30, 35.6, 33.3, 38.1],
        'series2' : [4, 5.5, 5.6, 6.3, 5.1],
        'series3' : [100, 100.5, 100.6, 100.2, 100.1]}
        df = pd.DataFrame(dataset).set_index('date')
        print(df)

        # Adjust Bayesian Ridge
        regr5 = linear_model.BayesianRidge()
        obj_DeepMTS = ns.DeepMTS(regr5, lags = 1, n_hidden_features=5)
        obj_DeepMTS.fit(df)
        print(obj_DeepMTS.predict())

        # with credible intervals
        print(obj_DeepMTS.predict(return_std=True, level=80))

        print(obj_DeepMTS.predict(return_std=True, level=95))
        ```

    """

    # construct the object -----

    def __init__(
        self,
        obj,
        n_layers=3,
        n_hidden_features=5,
        activation_name="relu",
        a=0.01,
        nodes_sim="sobol",
        bias=True,
        dropout=0,
        direct_link=True,
        n_clusters=2,
        cluster_encode=True,
        type_clust="kmeans",
        type_scaling=("std", "std", "std"),
        lags=1,
        type_pi="kde",
        replications=None,
        kernel=None,
        agg="mean",
        seed=123,
        backend="cpu",
        verbose=0,
        show_progress=True,
    ):
        assert int(lags) == lags, "parameter 'lags' should be an integer"
        assert n_layers >= 2, "must have n_layers >= 2"
        self.n_layers = int(n_layers)

        self.obj = DeepRegressor(
            obj=obj,
            verbose=0,
            n_layers=self.n_layers,
            n_hidden_features=n_hidden_features,
            activation_name=activation_name,
            a=a,
            nodes_sim=nodes_sim,
            bias=bias,
            dropout=dropout,
            direct_link=direct_link,
            n_clusters=n_clusters,
            cluster_encode=cluster_encode,
            type_clust=type_clust,
            type_scaling=type_scaling,
            seed=seed,
            backend=backend,
        )

        super().__init__(
            obj=self.obj,
            n_hidden_features=n_hidden_features,
            activation_name=activation_name,
            a=a,
            nodes_sim=nodes_sim,
            bias=bias,
            dropout=dropout,
            direct_link=direct_link,
            n_clusters=n_clusters,
            cluster_encode=cluster_encode,
            type_clust=type_clust,
            type_scaling=type_scaling,
            seed=seed,
            type_pi=type_pi,
            replications=replications,
            kernel=kernel,
            agg=agg,
            backend=backend,
            verbose=verbose,
            show_progress=show_progress,
        )



import functools


def memoize(func, maxsize=128):
    cache = func.cache = {}

    @functools.wraps(func)
    def memoized_func(*args, **kwargs):
        key = str(args) + str(kwargs)

        if key in cache:
            return cache[key]

        # else: key not in cache
        # you don't want it to grow indefinitely
        if len(cache) > maxsize:
            cache.clear()

        cache[key] = func(*args, **kwargs)
        return cache[key]

    return memoized_func



from .lazyClassifier import LazyClassifier
from .lazyRegressor import LazyRegressor
from .lazyMTS import LazyMTS
from .lazydeepClassifier import LazyDeepClassifier
from .lazydeepRegressor import LazyDeepRegressor
from .lazydeepMTS import LazyDeepMTS

__all__ = [
    "LazyClassifier",
    "LazyRegressor",
    "LazyMTS",
    "LazyDeepClassifier",
    "LazyDeepRegressor",
    "LazyDeepMTS",
]



from .icp import *

# TODO: move contents from nonconformist.icp here

# -----------------------------------------------------------------------------
# TcpClassifier
# -----------------------------------------------------------------------------
class TcpClassifier(BaseEstimator, ClassifierMixin):
	"""Transductive conformal classifier.

	Parameters
	----------
	nc_function : BaseScorer
		Nonconformity scorer object used to calculate nonconformity of
		calibration examples and test patterns. Should implement ``fit(x, y)``
		and ``calc_nc(x, y)``.

	smoothing : boolean
		Decides whether to use stochastic smoothing of p-values.

	Attributes
	----------
	train_x : numpy array of shape [n_cal_examples, n_features]
		Inputs of training set.

	train_y : numpy array of shape [n_cal_examples]
		Outputs of calibration set.

	nc_function : BaseScorer
		Nonconformity scorer object used to calculate nonconformity scores.

	classes : numpy array of shape [n_classes]
		List of class labels, with indices corresponding to output columns
		 of TcpClassifier.predict()

	See also
	--------
	IcpClassifier

	References
	----------
	.. [1] Vovk, V., Gammerman, A., & Shafer, G. (2005). Algorithmic learning
	in a random world. Springer Science & Business Media.

	Examples
	--------
	>>> import numpy as np
	>>> from sklearn.datasets import load_iris
	>>> from sklearn.svm import SVC
	>>> from nonconformist.base import ClassifierAdapter
	>>> from nonconformist.cp import TcpClassifier
	>>> from nonconformist.nc import ClassifierNc, MarginErrFunc
	>>> iris = load_iris()
	>>> idx = np.random.permutation(iris.target.size)
	>>> train = idx[:int(idx.size / 2)]
	>>> test = idx[int(idx.size / 2):]
	>>> model = ClassifierAdapter(SVC(probability=True))
	>>> nc = ClassifierNc(model, MarginErrFunc())
	>>> tcp = TcpClassifier(nc)
	>>> tcp.fit(iris.data[train, :], iris.target[train])
	>>> tcp.predict(iris.data[test, :], significance=0.10)
	...             # doctest: +SKIP
	array([[ True, False, False],
		[False,  True, False],
		...,
		[False,  True, False],
		[False,  True, False]], dtype=bool)
	"""

	def __init__(self, nc_function, condition=None, smoothing=True):
		self.train_x, self.train_y = None, None
		self.nc_function = nc_function
		super(TcpClassifier, self).__init__()

		# Check if condition-parameter is the default function (i.e.,
		# lambda x: 0). This is so we can safely clone the object without
		# the clone accidentally having self.conditional = True.
		default_condition = lambda x: 0
		is_default = (callable(condition) and
		              (condition.__code__.co_code ==
		               default_condition.__code__.co_code))

		if is_default:
			self.condition = condition
			self.conditional = False
		elif callable(condition):
			self.condition = condition
			self.conditional = True
		else:
			self.condition = lambda x: 0
			self.conditional = False

		self.smoothing = smoothing

		self.base_icp = IcpClassifier(
			self.nc_function,
			self.condition,
			self.smoothing
		)

		self.classes = None

	def fit(self, x, y):
		self.train_x, self.train_y = x, y
		self.classes = np.unique(y)

	def predict(self, x, significance=None):
		"""Predict the output values for a set of input patterns.

		Parameters
		----------
		x : numpy array of shape [n_samples, n_features]
			Inputs of patters for which to predict output values.

		significance : float or None
			Significance level (maximum allowed error rate) of predictions.
			Should be a float between 0 and 1. If ``None``, then the p-values
			are output rather than the predictions.

		Returns
		-------
		p : numpy array of shape [n_samples, n_classes]
			If significance is ``None``, then p contains the p-values for each
			sample-class pair; if significance is a float between 0 and 1, then
			p is a boolean array denoting which labels are included in the
			prediction sets.
		"""
		n_test = x.shape[0]
		n_train = self.train_x.shape[0]
		p = np.zeros((n_test, self.classes.size))
		for i in range(n_test):
			for j, y in enumerate(self.classes):
				train_x = np.vstack([self.train_x, x[i, :]])
				train_y = np.hstack([self.train_y, y])
				self.base_icp.fit(train_x, train_y)
				scores = self.base_icp.nc_function.score(train_x, train_y)
				ngt = (scores[:-1] > scores[-1]).sum()
				neq = (scores[:-1] == scores[-1]).sum()

				p[i, j] = calc_p(n_train, ngt, neq, self.smoothing)

		if significance is not None:
			return p > significance
		else:
			return p

	def predict_conf(self, x):
		"""Predict the output values for a set of input patterns, using
		the confidence-and-credibility output scheme.

		Parameters
		----------
		x : numpy array of shape [n_samples, n_features]
			Inputs of patters for which to predict output values.

		Returns
		-------
		p : numpy array of shape [n_samples, 3]
			p contains three columns: the first column contains the most
			likely class for each test pattern; the second column contains
			the confidence in the predicted class label, and the third column
			contains the credibility of the prediction.
		"""
		p = self.predict(x, significance=None)
		label = p.argmax(axis=1)
		credibility = p.max(axis=1)
		for i, idx in enumerate(label):
			p[i, idx] = -np.inf
		confidence = 1 - p.max(axis=1)

		return np.array([label, confidence, credibility]).T



import numpy as np
import pandas as pd
import sklearn.metrics as skm2
from copy import deepcopy
from tqdm import tqdm
from sklearn.base import ClassifierMixin
from ..custom import CustomClassifier
from ..utils import matrixops as mo


class DeepClassifier(CustomClassifier, ClassifierMixin):
    """
    Deep Classifier

    Parameters:

        verbose : int, optional (default=0)
            Monitor progress when fitting.

    Examples:

        ```python
        import nnetsauce as ns
        from sklearn.datasets import load_breast_cancer
        from sklearn.model_selection import train_test_split
        from sklearn.linear_model import LogisticRegressionCV
        data = load_breast_cancer()
        X = data.data
        y= data.target
        X_train, X_test, y_train, y_test = train_test_split(X, y, test_size=.2, random_state=123)
        obj = LogisticRegressionCV()
        clf = ns.DeepClassifier(obj)
        clf.fit(X_train, y_train)
        print(clf.score(clf.predict(X_test), y_test))
        ```
    """

    def __init__(
        self,
        obj,
        verbose=0,
        # Defining depth
        n_layers=3,
        # CustomClassifier attributes
        n_hidden_features=5,
        activation_name="relu",
        a=0.01,
        nodes_sim="sobol",
        bias=True,
        dropout=0,
        direct_link=True,
        n_clusters=2,
        cluster_encode=True,
        type_clust="kmeans",
        type_scaling=("std", "std", "std"),
        col_sample=1,
        row_sample=1,
        seed=123,
        backend="cpu",
    ):
        super().__init__(
            obj=obj,
            n_hidden_features=n_hidden_features,
            activation_name=activation_name,
            a=a,
            nodes_sim=nodes_sim,
            bias=bias,
            dropout=dropout,
            direct_link=direct_link,
            n_clusters=n_clusters,
            cluster_encode=cluster_encode,
            type_clust=type_clust,
            type_scaling=type_scaling,
            col_sample=col_sample,
            row_sample=row_sample,
            seed=seed,
            backend=backend,
        )

        assert n_layers >= 2, "must have n_layers >= 2"

        self.stacked_obj = obj
        self.verbose = verbose
        self.n_layers = n_layers

    def fit(self, X, y):
        """Fit Classification algorithms to X and y.
        Parameters
        ----------
        X : array-like,
            Training vectors, where rows is the number of samples
            and columns is the number of features.
        y : array-like,
            Training vectors, where rows is the number of samples
            and columns is the number of features.
        Returns
        -------
        A fitted object
        """

        if isinstance(X, np.ndarray):
            X = pd.DataFrame(X)

        self.n_classes_ = len(np.unique(y)) # for compatibility with sklearn 
        
        # init layer
        self.stacked_obj = CustomClassifier(
            obj=self.stacked_obj,
            n_hidden_features=self.n_hidden_features,
            activation_name=self.activation_name,
            a=self.a,
            nodes_sim=self.nodes_sim,
            bias=self.bias,
            dropout=self.dropout,
            direct_link=self.direct_link,
            n_clusters=self.n_clusters,
            cluster_encode=self.cluster_encode,
            type_clust=self.type_clust,
            type_scaling=self.type_scaling,
            col_sample=self.col_sample,
            row_sample=self.row_sample,
            seed=self.seed,
            backend=self.backend,
        )

        self.stacked_obj.fit(X, y)

        if self.verbose > 0:
            iterator = tqdm(range(self.n_layers - 1))
        else:
            iterator = range(self.n_layers - 1)

        for _ in iterator:
            self.stacked_obj = deepcopy(
                CustomClassifier(
                    obj=self.stacked_obj,
                    n_hidden_features=self.n_hidden_features,
                    activation_name=self.activation_name,
                    a=self.a,
                    nodes_sim=self.nodes_sim,
                    bias=self.bias,
                    dropout=self.dropout,
                    direct_link=self.direct_link,
                    n_clusters=self.n_clusters,
                    cluster_encode=self.cluster_encode,
                    type_clust=self.type_clust,
                    type_scaling=self.type_scaling,
                    col_sample=self.col_sample,
                    row_sample=self.row_sample,
                    seed=self.seed,
                    backend=self.backend,
                )
            )

            # self.stacked_obj.fit(X, y)

        self.stacked_obj.fit(X, y)

        self.obj = deepcopy(self.stacked_obj)

        return self.obj

    def predict(self, X):
        return self.obj.predict(X)

    def predict_proba(self, X):
        return self.obj.predict_proba(X)

    def score(self, X, y, scoring=None):
        return self.obj.score(X, y, scoring)



import numpy as np
import pandas as pd 

def getsims(input_data, ix):

    if isinstance(input_data[0], pd.DataFrame): # kde
        n_sims = len(input_data)
        res = [input_data[i].iloc[:, ix].values for i in range(n_sims)] 
        return np.asarray(res).T       

    if isinstance(input_data[0], tuple): # GP posterior       
        h = len(input_data)        
        n_sims = len(input_data[0][0][0])
        res = [[input_data[hx][ix][0][i] for i in range(n_sims)] for hx in range(h)]
        return np.asarray(res)

    


# Authors: T. Moudiki
#
# License: BSD 3 Clause Clear

import numpy as np
from tqdm import tqdm
from ..sampling.helpers import dosubsample


# 0 - utils -----

# 0 - 0 data structures & funcs -----


def call_f(f, x):
    return f(x)


def calc_grad(f, x):
    return numerical_gradient(f, x)


def calc_hessian(f, x):
    return numerical_hessian(f, x)


def generate_index(response, batch_prop=1.0, randomization="strat", seed=123):
    """Generation of indices for Stochastic gradient descent."""

    n = len(response)

    if batch_prop < 1:
        if randomization == "strat":
            return dosubsample(response, batch_prop, seed=seed)

        if randomization == "shuffle":
            np.random.seed(seed)
            return np.asarray(
                np.floor(
                    np.random.choice(
                        range(n), size=np.floor(batch_prop * n), replace=False
                    )
                ),
                dtype=np.int,
            )
    return None


def check_is_min(f, x):
    return (numerical_hessian(f, x)[1]) * 1


def calc_fplus(x, alpha, p_k):
    p = len(x)
    res = np.zeros(p)
    j = 0

    for j in range(p):
        res[j] = x[j] + alpha * p_k[j]

    return res


# Useful functions -----

# 1 - gradient-----


def numerical_gradient(f, x, **kwargs):
    p = len(x)
    ix = 0
    eps_factor = 6.055454452393343e-06  # machine eps ** (1/3)
    zero = 2.220446049250313e-16
    res = np.zeros_like(x)

    def f_(x):
        return f(x, **kwargs)

    for ix in range(p):
        value_x = x[ix]

        h = max(eps_factor * value_x, 1e-8)

        x[ix] = value_x + h
        fx_plus = call_f(f_, x)

        x[ix] = value_x - h
        fx_minus = call_f(f_, x)

        x[ix] = value_x  # restore

        res[ix] = (fx_plus - fx_minus) / (2 * h)

    return np.asarray(res)


# 2 - hessian-----


def numerical_hessian(f, x, **kwargs):
    p = len(x)
    ix = 0
    jx = 0
    eps_factor = 0.0001220703125  # machine eps ** (1/4)
    zero = 2.220446049250313e-16
    H = np.zeros((p, p))
    temp = 0

    def f_(x):
        return f(x, **kwargs)

    fx = call_f(f_, x)

    for ix in range(p):
        for jx in range(ix, p):
            if ix < jx:
                value_x = x[ix]
                value_y = x[jx]

                h = max(eps_factor * value_x, 1e-8)
                k = max(eps_factor * value_y, 1e-8)

                x[ix] = value_x + h
                x[jx] = value_y + k
                fx_plus = call_f(f_, x)

                x[ix] = value_x + h
                x[jx] = value_y - k
                fx_plus_minus = call_f(f_, x)

                x[ix] = value_x - h
                x[jx] = value_y + k
                fx_minus_plus = call_f(f_, x)

                x[ix] = value_x - h
                x[jx] = value_y - k
                fx_minus = call_f(f_, x)

                x[ix] = value_x  # restore
                x[jx] = value_y  # restore

                temp = (fx_plus - fx_plus_minus - fx_minus_plus + fx_minus) / (
                    4 * h * k
                )

            else:
                value_x = x[ix]

                h = max(eps_factor * value_x, 1e-8)

                x[ix] = value_x + h
                fx_plus = call_f(f_, x)

                x[ix] = value_x - h
                fx_minus = call_f(f_, x)

                x[ix] = value_x  # restore

                temp = (fx_plus - 2 * fx + fx_minus) / (h**2)

            H[ix, jx] = temp
            H[jx, ix] = temp

    res = np.asarray(H)

    return res, all(np.linalg.eig(res)[0] > 0)


# 3 - One-hot encoder -----


def one_hot_encode(y, n_classes):
    n_obs = len(y)
    res = np.zeros((n_obs, n_classes), dtype=float)

    for i in range(n_obs):
        res[i, y[i]] = 1

    return np.asarray(res)


# Coordinate descent (Stochastic) -----

# 1 - algos -----


def scd(
    loss_func,
    response,
    x,
    num_iters=200,
    batch_prop=1.0,
    learning_rate=0.01,
    mass=0.9,
    decay=0.1,
    method="momentum",
    randomization="strat",
    tolerance=1e-3,
    verbose=1,
    **kwargs,
):
    """Stochastic coordinate descent with momentum and adaptive learning rates."""

    i = 0
    j = 0
    n = len(response)
    p = len(x)
    velocity = np.zeros(p)
    losses = []

    if verbose == 1:
        iterator = tqdm(range(num_iters))
    else:
        iterator = range(num_iters)

    def f(x):
        return loss_func(x, **kwargs)

    if method == "momentum":
        for i in iterator:
            idx = generate_index(
                response=response,
                batch_prop=batch_prop,
                randomization=randomization,
                seed=i,
            )

            def f_j(h, xx, j):
                value_x = 0
                res = 0
                value_x = xx[j]
                xx[j] = xx[j] + h
                res = loss_func(xx, row_index=idx, **kwargs)
                xx[j] = value_x
                return res

            diff = -np.asarray(x)

            if verbose == 2:
                print(f"\n x prev: {np.asarray(x)}")

            for j in range(p):
                h0 = 6.055454452393343e-06 * x[j]
                grad_x = (f_j(h0, x, j) - f_j(-h0, x, j)) / (2 * h0)
                velocity[j] = mass * velocity[j] - learning_rate * grad_x
                x[j] = x[j] + velocity[j]

            diff += np.asarray(x)

            if verbose == 2:
                print(f"\n x new: {np.asarray(x)}")

            losses.append(f(x))

            if (len(losses) > 3) and (
                np.abs(np.diff(losses[-2:])[0]) < tolerance
            ):
                break

            if verbose == 2:
                print("\n")
                print(f"iter {i+1} - decrease -----")

                try:
                    print(np.linalg.norm(diff, 1))
                except:
                    pass

                print(f"iter {i+1} - loss -----")
                print(np.flip(losses)[0])

    if method in ("exp", "poly"):
        for i in iterator:
            idx = generate_index(
                response=response,
                batch_prop=batch_prop,
                randomization=randomization,
                seed=i,
            )

            def f_j(h, xx, j):
                value_x = 0
                res = 0
                value_x = xx[j]
                xx[j] = xx[j] + h
                res = loss_func(xx, row_index=idx, **kwargs)
                xx[j] = value_x
                return res

            decay_rate = (
                (1 + decay * i) if (method == "poly") else np.exp(decay * i)
            )

            diff = -np.asarray(x)

            if verbose == 2:
                print(f"\n x prev: {np.asarray(x)}")

            losses.append(f(x))

            for j in range(p):
                h0 = 6.055454452393343e-06 * x[j]
                grad_x = (f_j(h0, x, j) - f_j(-h0, x, j)) / (2 * h0)
                x[j] = x[j] - grad_x * learning_rate / decay_rate

            diff += np.asarray(x)

            if verbose == 2:
                print(f"\n x new: {np.asarray(x)}")

            if (len(losses) > 3) and (
                np.abs(np.diff(losses[-2:])[0]) < tolerance
            ):
                break

            if verbose == 2:
                print("\n")
                print(f"iter {i+1} - decrease -----")

                try:
                    print(np.linalg.norm(diff, 1))
                except:
                    pass

                print(f"iter {i+1} - loss -----")
                print(np.flip(losses)[0])

    return np.asarray(x), num_iters, losses


# Gradient descent (Stochastic) -----


def sgd(
    loss_func,
    response,
    x,
    num_iters=200,
    batch_prop=1.0,
    learning_rate=0.01,
    mass=0.9,
    decay=0.1,
    method="momentum",
    randomization="strat",
    tolerance=1e-3,
    verbose=1,
    **kwargs,
):
    """Stochastic gradient descent with momentum and adaptive learning rates."""

    i = 0
    j = 0
    n = len(response)
    p = len(x)
    velocity = np.zeros(p)
    grad_i = np.zeros(p)
    losses = []

    if verbose == 1:
        iterator = tqdm(range(num_iters))
    else:
        iterator = range(num_iters)

    def f(x):
        return loss_func(x, **kwargs)

    if method == "momentum":
        for i in iterator:
            idx = generate_index(
                response=response,
                batch_prop=batch_prop,
                randomization=randomization,
                seed=i,
            )

            def objective(x):
                return loss_func(x, row_index=idx, **kwargs)

            # grad_i = numerical_gradient(objective, x)
            grad_i = calc_grad(objective, x)

            diff = -np.asarray(x)

            if verbose == 2:
                print(f"\n x prev: {np.asarray(x)}")

            for j in range(p):
                velocity[j] = mass * velocity[j] - learning_rate * grad_i[j]
                x[j] = x[j] + velocity[j]

            diff += np.asarray(x)

            if verbose == 2:
                print(f"\n x new: {np.asarray(x)}")

            losses.append(f(x))

            if (len(losses) > 3) and (
                np.abs(np.diff(losses[-2:])[0]) < tolerance
            ):
                break

            if verbose == 2:
                print("\n")
                print(f"iter {i+1} - decrease -----")

                try:
                    print(np.linalg.norm(diff, 1))
                except:
                    pass

                print(f"iter {i+1} - loss -----")
                print(np.flip(losses)[0])

    if method in ("exp", "poly"):
        for i in iterator:
            idx = generate_index(
                response=response,
                batch_prop=batch_prop,
                randomization=randomization,
                seed=i,
            )

            def objective(x):
                return loss_func(x, row_index=idx, **kwargs)

            # grad_i = numerical_gradient(objective, x)
            grad_i = calc_grad(objective, x)

            decay_rate = (
                (1 + decay * i) if (method == "poly") else np.exp(decay * i)
            )

            diff = -np.asarray(x)

            if verbose == 2:
                print(f"\n x prev: {np.asarray(x)}")

            for j in range(p):
                x[j] = x[j] - grad_i[j] * learning_rate / decay_rate

            if verbose == 2:
                print(f"\n x new: {np.asarray(x)}")

            diff += np.asarray(x)

            losses.append(f(x))

            if (len(losses) > 3) and (
                np.abs(np.diff(losses[-2:])[0]) < tolerance
            ):
                break

            if verbose == 2:
                print("\n")
                print(f"iter {i+1} - decrease -----")

                try:
                    print(np.linalg.norm(diff, 1))
                except:
                    pass

                print(f"iter {i+1} - loss -----")
                print(np.flip(losses)[0])

    return np.asarray(x), num_iters, losses



import numpy as np
import pandas as pd
from tqdm import tqdm
import time
from sklearn.utils.discovery import all_estimators
from sklearn.pipeline import Pipeline
from sklearn.impute import SimpleImputer
from sklearn.preprocessing import StandardScaler, OneHotEncoder, OrdinalEncoder
from sklearn.compose import ColumnTransformer
from sklearn.base import RegressorMixin
from sklearn.metrics import (
    r2_score,
    mean_squared_error,
    mean_absolute_error,
    mean_pinball_loss,
    mean_absolute_percentage_error,
)
from .config import REGRESSORSMTS
from ..mts import MTS
from ..utils import convert_df_to_numeric

import warnings

warnings.filterwarnings("ignore")
pd.set_option("display.precision", 2)
pd.set_option("display.float_format", lambda x: "%.2f" % x)

numeric_transformer = Pipeline(
    steps=[
        ("imputer", SimpleImputer(strategy="mean")),
        ("scaler", StandardScaler()),
    ]
)

categorical_transformer_low = Pipeline(
    steps=[
        ("imputer", SimpleImputer(strategy="constant", fill_value="missing")),
        (
            "encoding",
            OneHotEncoder(handle_unknown="ignore", sparse_output=False),
        ),
    ]
)

categorical_transformer_high = Pipeline(
    steps=[
        ("imputer", SimpleImputer(strategy="constant", fill_value="missing")),
        # 'OrdinalEncoder' Raise a ValueError when encounters an unknown value. Check https://github.com/scikit-learn/scikit-learn/pull/13423
        ("encoding", OrdinalEncoder()),
    ]
)


# Helper function


def get_card_split(df, cols, n=11):
    cond = df[cols].nunique() > n
    card_high = cols[cond]
    card_low = cols[~cond]
    return card_low, card_high


def adjusted_rsquared(r2, n, p):
    return 1 - (1 - r2) * ((n - 1) / (n - p - 1))


# def mean_percentage_error(y_pred, y_true):
#     pe = (y_pred - y_true) / np.maximum(np.abs(y_true),
#                                         np.finfo(np.float64).eps)
#     return np.average(pe)


class LazyMTS(MTS):
    """

    Fitting -- almost -- all the regression algorithms with layers of
    nnetsauce's CustomRegressor to multivariate time series
    and returning their scores.

    Parameters:

        verbose: int, optional (default=0)
            Any positive number for verbosity.

        ignore_warnings: bool, optional (default=True)
            When set to True, the warning related to algorigms that are not
            able to run are ignored.

        custom_metric: function, optional (default=None)
            When function is provided, models are evaluated based on the custom
              evaluation metric provided.

        predictions: bool, optional (default=False)
            When set to True, the predictions of all the models models are returned as dataframe.

        sort_by: string, optional (default='RMSE')
            Sort models by a metric. Available options are 'RMSE', 'MAE', 'MPL', 'MPE', 'MAPE',
            'R-Squared', 'Adjusted R-Squared' or a custom metric identified by its name and
            provided by custom_metric.

        random_state: int, optional (default=42)
            Reproducibiility seed.

        estimators: list, optional (default='all')
            list of Estimators (regression algorithms) names or just 'all' (default='all')

        preprocess: bool, preprocessing is done when set to True

        n_jobs : int, when possible, run in parallel
            For now, only used by individual models that support it.

        n_layers: int, optional (default=3)
            Number of layers of CustomRegressors to be used.

        All the other parameters are the same as MTS's.

    Examples:

        See https://thierrymoudiki.github.io/blog/2023/10/29/python/quasirandomizednn/MTS-LazyPredict

    """

    def __init__(
        self,
        verbose=0,
        ignore_warnings=True,
        custom_metric=None,
        predictions=False,
        random_state=42,
        estimators="all",
        preprocess=False,
        # MTS attributes
        obj=None,
        n_hidden_features=5,
        activation_name="relu",
        a=0.01,
        nodes_sim="sobol",
        bias=True,
        dropout=0,
        direct_link=True,
        n_clusters=2,
        cluster_encode=True,
        type_clust="kmeans",
        type_scaling=("std", "std", "std"),
        lags=1,
        replications=None,
        kernel=None,
        agg="mean",
        seed=123,
        backend="cpu",
        show_progress=False,
    ):
        self.verbose = verbose
        self.ignore_warnings = ignore_warnings
        self.custom_metric = custom_metric
        self.predictions = predictions
        self.models = {}
        self.random_state = random_state
        self.estimators = estimators
        self.preprocess = preprocess
        super().__init__(
            obj=obj,
            n_hidden_features=n_hidden_features,
            activation_name=activation_name,
            a=a,
            nodes_sim=nodes_sim,
            bias=bias,
            dropout=dropout,
            direct_link=direct_link,
            n_clusters=n_clusters,
            cluster_encode=cluster_encode,
            type_clust=type_clust,
            type_scaling=type_scaling,
            seed=seed,
            backend=backend,
            lags=lags,
            replications=replications,
            kernel=kernel,
            agg=agg,
            show_progress=show_progress,
        )

    def fit(self, X_train, X_test, xreg=None, **kwargs):
        """Fit Regression algorithms to X_train, predict and score on X_test.

        Parameters:

            X_train: array-like,
                Training vectors, where rows is the number of samples
                and columns is the number of features.

            X_test: array-like,
                Testing vectors, where rows is the number of samples
                and columns is the number of features.

            xreg: array-like, optional (default=None)
                Additional (external) regressors to be passed to self.obj
                xreg must be in 'increasing' order (most recent observations last)

        Returns:

            scores: Pandas DataFrame
                Returns metrics of all the models in a Pandas DataFrame.

            predictions: Pandas DataFrame
                Returns predictions of all the models in a Pandas DataFrame.

        """
        R2 = []
        ADJR2 = []
        ME = []
        MPL = []
        RMSE = []
        MAE = []
        MPE = []
        MAPE = []

        # WIN = []
        names = []
        TIME = []
        predictions = {}

        if self.custom_metric:
            CUSTOM_METRIC = []

        if isinstance(X_train, np.ndarray):
            X_train = pd.DataFrame(X_train)
            X_test = pd.DataFrame(X_test)

        X_train = convert_df_to_numeric(X_train)
        X_test = convert_df_to_numeric(X_test)

        numeric_features = X_train.select_dtypes(include=[np.number]).columns
        categorical_features = X_train.select_dtypes(include=["object"]).columns

        categorical_low, categorical_high = get_card_split(
            X_train, categorical_features
        )

        if self.preprocess:
            preprocessor = ColumnTransformer(
                transformers=[
                    ("numeric", numeric_transformer, numeric_features),
                    (
                        "categorical_low",
                        categorical_transformer_low,
                        categorical_low,
                    ),
                    (
                        "categorical_high",
                        categorical_transformer_high,
                        categorical_high,
                    ),
                ]
            )

        if self.estimators == "all":
            self.regressors = REGRESSORSMTS
        else:
            self.regressors = [
                ("MTS(" + est[0] + ")", est[1])
                for est in all_estimators()
                if (
                    issubclass(est[1], RegressorMixin)
                    and (est[0] in self.estimators)
                )
            ]

        if self.preprocess is True:
            for name, model in tqdm(self.regressors):  # do parallel exec
                start = time.time()
                try:
                    if "random_state" in model().get_params().keys():
                        pipe = Pipeline(
                            steps=[
                                ("preprocessor", preprocessor),
                                (
                                    "regressor",
                                    MTS(
                                        obj=model(
                                            random_state=self.random_state,
                                            **kwargs
                                        ),
                                        n_hidden_features=self.n_hidden_features,
                                        activation_name=self.activation_name,
                                        a=self.a,
                                        nodes_sim=self.nodes_sim,
                                        bias=self.bias,
                                        dropout=self.dropout,
                                        direct_link=self.direct_link,
                                        n_clusters=self.n_clusters,
                                        cluster_encode=self.cluster_encode,
                                        type_clust=self.type_clust,
                                        type_scaling=self.type_scaling,
                                        lags=self.lags,
                                        replications=self.replications,
                                        kernel=self.kernel,
                                        agg=self.agg,
                                        seed=self.seed,
                                        backend=self.backend,
                                        show_progress=self.show_progress,
                                    ),
                                ),
                            ]
                        )
                    else:
                        pipe = Pipeline(
                            steps=[
                                ("preprocessor", preprocessor),
                                (
                                    "regressor",
                                    MTS(
                                        obj=model(**kwargs),
                                        n_hidden_features=self.n_hidden_features,
                                        activation_name=self.activation_name,
                                        a=self.a,
                                        nodes_sim=self.nodes_sim,
                                        bias=self.bias,
                                        dropout=self.dropout,
                                        direct_link=self.direct_link,
                                        n_clusters=self.n_clusters,
                                        cluster_encode=self.cluster_encode,
                                        type_clust=self.type_clust,
                                        type_scaling=self.type_scaling,
                                        lags=self.lags,
                                        replications=self.replications,
                                        kernel=self.kernel,
                                        agg=self.agg,
                                        seed=self.seed,
                                        backend=self.backend,
                                        show_progress=self.show_progress,
                                    ),
                                ),
                            ]
                        )

                    pipe.fit(X_train, **kwargs)
                    # pipe.fit(X_train, xreg=xreg)

                    self.models[name] = pipe

                    X_pred = pipe["regressor"].predict(
                        h=X_test.shape[0], **kwargs
                    )

                    rmse = mean_squared_error(X_test, X_pred, squared=False)
                    mae = mean_absolute_error(X_test, X_pred)
                    mpl = mean_pinball_loss(X_test, X_pred)

                    names.append(name)
                    RMSE.append(rmse)
                    MAE.append(mae)
                    MPL.append(mpl)
                    TIME.append(time.time() - start)

                    if self.custom_metric:
                        custom_metric = self.custom_metric(X_test, X_pred)
                        CUSTOM_METRIC.append(custom_metric)

                    if self.verbose > 0:
                        scores_verbose = {
                            "Model": name,
                            # "R-Squared": r_squared,
                            # "Adjusted R-Squared": adj_rsquared,
                            "RMSE": rmse,
                            "MAE": mae,
                            "MPL": mpl,
                            # "MPE": mpe,
                            # "MAPE": mape,
                            "Time taken": time.time() - start,
                        }

                        if self.custom_metric:
                            scores_verbose[self.custom_metric.__name__] = (
                                custom_metric
                            )

                        print(scores_verbose)
                    if self.predictions:
                        predictions[name] = X_pred
                except Exception as exception:
                    if self.ignore_warnings is False:
                        print(name + " model failed to execute")
                        print(exception)

        else:  # no preprocessing
            for name, model in tqdm(self.regressors):  # do parallel exec
                start = time.time()
                try:
                    if "random_state" in model().get_params().keys():
                        pipe = MTS(
                            obj=model(random_state=self.random_state, **kwargs),
                            n_hidden_features=self.n_hidden_features,
                            activation_name=self.activation_name,
                            a=self.a,
                            nodes_sim=self.nodes_sim,
                            bias=self.bias,
                            dropout=self.dropout,
                            direct_link=self.direct_link,
                            n_clusters=self.n_clusters,
                            cluster_encode=self.cluster_encode,
                            type_clust=self.type_clust,
                            type_scaling=self.type_scaling,
                            lags=self.lags,
                            replications=self.replications,
                            kernel=self.kernel,
                            agg=self.agg,
                            seed=self.seed,
                            backend=self.backend,
                            show_progress=self.show_progress,
                        )
                    else:
                        pipe = MTS(
                            obj=model(**kwargs),
                            n_hidden_features=self.n_hidden_features,
                            activation_name=self.activation_name,
                            a=self.a,
                            nodes_sim=self.nodes_sim,
                            bias=self.bias,
                            dropout=self.dropout,
                            direct_link=self.direct_link,
                            n_clusters=self.n_clusters,
                            cluster_encode=self.cluster_encode,
                            type_clust=self.type_clust,
                            type_scaling=self.type_scaling,
                            lags=self.lags,
                            replications=self.replications,
                            kernel=self.kernel,
                            agg=self.agg,
                            seed=self.seed,
                            backend=self.backend,
                            show_progress=self.show_progress,
                        )

                    pipe.fit(X_train, xreg, **kwargs)
                    # pipe.fit(X_train, xreg=xreg) # DO xreg like in `ahead`

                    self.models[name] = pipe

                    if self.preprocess is True:
                        X_pred = pipe["regressor"].predict(
                            h=X_test.shape[0], **kwargs
                        )
                    else:
                        X_pred = pipe.predict(
                            h=X_test.shape[0], **kwargs
                        )  # X_pred = pipe.predict(h=X_test.shape[0], new_xreg=new_xreg) ## DO xreg like in `ahead`

                    rmse = mean_squared_error(X_test, X_pred, squared=False)
                    mae = mean_absolute_error(X_test, X_pred)
                    mpl = mean_pinball_loss(X_test, X_pred)

                    names.append(name)
                    RMSE.append(rmse)
                    MAE.append(mae)
                    MPL.append(mpl)
                    TIME.append(time.time() - start)

                    if self.custom_metric:
                        custom_metric = self.custom_metric(X_test, X_pred)
                        CUSTOM_METRIC.append(custom_metric)

                    if self.verbose > 0:
                        scores_verbose = {
                            "Model": name,
                            # "R-Squared": r_squared,
                            # "Adjusted R-Squared": adj_rsquared,
                            "RMSE": rmse,
                            "MAE": mae,
                            "MPL": mpl,
                            # "MPE": mpe,
                            # "MAPE": mape,
                            "Time taken": time.time() - start,
                        }

                        if self.custom_metric:
                            scores_verbose[self.custom_metric.__name__] = (
                                custom_metric
                            )

                        print(scores_verbose)
                    if self.predictions:
                        predictions[name] = X_pred
                except Exception as exception:
                    if self.ignore_warnings is False:
                        print(name + " model failed to execute")
                        print(exception)

        scores = {
            "Model": names,
            # "Adjusted R-Squared": ADJR2,
            # "R-Squared": R2,
            "RMSE": RMSE,
            "MAE": MAE,
            "MPL": MPL,
            # "MPE": MPE,
            # "MAPE": MAPE,
            "Time Taken": TIME,
        }

        if self.custom_metric:
            scores[self.custom_metric.__name__] = CUSTOM_METRIC

        scores = pd.DataFrame(scores)
        scores = scores.sort_values(by="RMSE", ascending=True).set_index(
            "Model"
        )

        if self.predictions:
            predictions_df = pd.DataFrame.from_dict(predictions)
        return scores, predictions_df if self.predictions is True else scores

    def provide_models(self, X_train, X_test):
        """
        This function returns all the model objects trained in fit function.
        If fit is not called already, then we call fit and then return the models.

        Parameters:

            X_train : array-like,
                Training vectors, where rows is the number of samples
                and columns is the number of features.

            X_test : array-like,
                Testing vectors, where rows is the number of samples
                and columns is the number of features.

        Returns:

            models: dict-object,
                Returns a dictionary with each model pipeline as value
                with key as name of models.

        """
        if len(self.models.keys()) == 0:
            self.fit(X_train, X_test)

        return self.models



#!/usr/bin/env python

"""
Evaluation of conformal predictors.
"""

# Authors: Henrik Linusson

# TODO: cross_val_score/run_experiment should possibly allow multiple to be evaluated on identical folding

from __future__ import division

from nonconformist.base import RegressorMixin, ClassifierMixin

import sys
import numpy as np
import pandas as pd

from sklearn.cross_validation import StratifiedShuffleSplit
from sklearn.cross_validation import KFold
from sklearn.cross_validation import train_test_split
from sklearn.base import clone, BaseEstimator


class BaseIcpCvHelper(BaseEstimator):
	"""Base class for cross validation helpers.
	"""
	def __init__(self, icp, calibration_portion):
		super(BaseIcpCvHelper, self).__init__()
		self.icp = icp
		self.calibration_portion = calibration_portion

	def predict(self, x, significance=None):
			return self.icp.predict(x, significance)


class ClassIcpCvHelper(BaseIcpCvHelper, ClassifierMixin):
	"""Helper class for running the ``cross_val_score`` evaluation
	method on IcpClassifiers.

	See also
	--------
	IcpRegCrossValHelper

	Examples
	--------
	>>> from sklearn.datasets import load_iris
	>>> from sklearn.ensemble import RandomForestClassifier
	>>> from nonconformist.icp import IcpClassifier
	>>> from nonconformist.nc import ClassifierNc, MarginErrFunc
	>>> from nonconformist.evaluation import ClassIcpCvHelper
	>>> from nonconformist.evaluation import class_mean_errors
	>>> from nonconformist.evaluation import cross_val_score
	>>> data = load_iris()
	>>> nc = ProbEstClassifierNc(RandomForestClassifier(), MarginErrFunc())
	>>> icp = IcpClassifier(nc)
	>>> icp_cv = ClassIcpCvHelper(icp)
	>>> cross_val_score(icp_cv,
	...                 data.data,
	...                 data.target,
	...                 iterations=2,
	...                 folds=2,
	...                 scoring_funcs=[class_mean_errors],
	...                 significance_levels=[0.1])
	...     # doctest: +SKIP
	   class_mean_errors  fold  iter  significance
	0           0.013333     0     0           0.1
	1           0.080000     1     0           0.1
	2           0.053333     0     1           0.1
	3           0.080000     1     1           0.1
	"""
	def __init__(self, icp, calibration_portion=0.25):
		super(ClassIcpCvHelper, self).__init__(icp, calibration_portion)

	def fit(self, x, y):
		split = StratifiedShuffleSplit(y, n_iter=1,
		                               test_size=self.calibration_portion)
		for train, cal in split:
			self.icp.fit(x[train, :], y[train])
			self.icp.calibrate(x[cal, :], y[cal])


class RegIcpCvHelper(BaseIcpCvHelper, RegressorMixin):
	"""Helper class for running the ``cross_val_score`` evaluation
	method on IcpRegressors.

	See also
	--------
	IcpClassCrossValHelper

	Examples
	--------
	>>> from sklearn.datasets import load_boston
	>>> from sklearn.ensemble import RandomForestRegressor
	>>> from nonconformist.icp import IcpRegressor
	>>> from nonconformist.nc import RegressorNc, AbsErrorErrFunc
	>>> from nonconformist.evaluation import RegIcpCvHelper
	>>> from nonconformist.evaluation import reg_mean_errors
	>>> from nonconformist.evaluation import cross_val_score
	>>> data = load_boston()
	>>> nc = RegressorNc(RandomForestRegressor(), AbsErrorErrFunc())
	>>> icp = IcpRegressor(nc)
	>>> icp_cv = RegIcpCvHelper(icp)
	>>> cross_val_score(icp_cv,
	...                 data.data,
	...                 data.target,
	...                 iterations=2,
	...                 folds=2,
	...                 scoring_funcs=[reg_mean_errors],
	...                 significance_levels=[0.1])
	...     # doctest: +SKIP
	   fold  iter  reg_mean_errors  significance
	0     0     0         0.185771           0.1
	1     1     0         0.138340           0.1
	2     0     1         0.071146           0.1
	3     1     1         0.043478           0.1
	"""
	def __init__(self, icp, calibration_portion=0.25):
		super(RegIcpCvHelper, self).__init__(icp, calibration_portion)

	def fit(self, x, y):
		split = train_test_split(x, y, test_size=self.calibration_portion)
		x_tr, x_cal, y_tr, y_cal = split[0], split[1], split[2], split[3]
		self.icp.fit(x_tr, y_tr)
		self.icp.calibrate(x_cal, y_cal)


# -----------------------------------------------------------------------------
#
# -----------------------------------------------------------------------------
def cross_val_score(model,x, y, iterations=10, folds=10, fit_params=None,
					scoring_funcs=None, significance_levels=None,
					verbose=False):
	"""Evaluates a conformal predictor using cross-validation.

	Parameters
	----------
	model : object
		Conformal predictor to evaluate.

	x : numpy array of shape [n_samples, n_features]
		Inputs of data to use for evaluation.

	y : numpy array of shape [n_samples]
		Outputs of data to use for evaluation.

	iterations : int
		Number of iterations to use for evaluation. The data set is randomly
		shuffled before each iteration.

	folds : int
		Number of folds to use for evaluation.

	fit_params : dictionary
		Parameters to supply to the conformal prediction object on training.

	scoring_funcs : iterable
		List of evaluation functions to apply to the conformal predictor in each
		fold. Each evaluation function should have a signature
		``scorer(prediction, y, significance)``.

	significance_levels : iterable
		List of significance levels at which to evaluate the conformal
		predictor.

	verbose : boolean
		Indicates whether to output progress information during evaluation.

	Returns
	-------
	scores : pandas DataFrame
		Tabulated results for each iteration, fold and evaluation function.
	"""

	fit_params = fit_params if fit_params else {}
	significance_levels = (significance_levels if significance_levels
	                       is not None else np.arange(0.01, 1.0, 0.01))

	df = pd.DataFrame()

	columns = ['iter',
			   'fold',
			   'significance',
			   ] + [f.__name__ for f in scoring_funcs]
	for i in range(iterations):
		idx = np.random.permutation(y.size)
		x, y = x[idx, :], y[idx]
		cv = KFold(y.size, folds)
		for j, (train, test) in enumerate(cv):
			if verbose:
				sys.stdout.write('\riter {}/{} fold {}/{}'.format(
					i + 1,
					iterations,
					j + 1,
					folds
				))
			m = clone(model)
			m.fit(x[train, :], y[train], **fit_params)
			prediction = m.predict(x[test, :], significance=None)
			for k, s in enumerate(significance_levels):
				scores = [scoring_func(prediction, y[test], s)
						  for scoring_func in scoring_funcs]
				df_score = pd.DataFrame([[i, j, s] + scores],
											columns=columns)
				df = df.append(df_score, ignore_index=True)

	return df


def run_experiment(models, csv_files, iterations=10, folds=10, fit_params=None,
				   scoring_funcs=None, significance_levels=None,
				   normalize=False, verbose=False, header=0):
	"""Performs a cross-validation evaluation of one or several conformal
	predictors on a	collection of data sets in csv format.

	Parameters
	----------
	models : object or iterable
		Conformal predictor(s) to evaluate.

	csv_files : iterable
		List of file names (with absolute paths) containing csv-data, used to
		evaluate the conformal predictor.

	iterations : int
		Number of iterations to use for evaluation. The data set is randomly
		shuffled before each iteration.

	folds : int
		Number of folds to use for evaluation.

	fit_params : dictionary
		Parameters to supply to the conformal prediction object on training.

	scoring_funcs : iterable
		List of evaluation functions to apply to the conformal predictor in each
		fold. Each evaluation function should have a signature
		``scorer(prediction, y, significance)``.

	significance_levels : iterable
		List of significance levels at which to evaluate the conformal
		predictor.

	verbose : boolean
		Indicates whether to output progress information during evaluation.

	Returns
	-------
	scores : pandas DataFrame
		Tabulated results for each data set, iteration, fold and
		evaluation function.
	"""
	df = pd.DataFrame()
	if not hasattr(models, '__iter__'):
		models = [models]

	for model in models:
		is_regression = model.get_problem_type() == 'regression'

		n_data_sets = len(csv_files)
		for i, csv_file in enumerate(csv_files):
			if verbose:
				print('\n{} ({} / {})'.format(csv_file, i + 1, n_data_sets))
			data = pd.read_csv(csv_file, header=header)
			x, y = data.values[:, :-1], data.values[:, -1]
			x = np.array(x, dtype=np.float64)
			if normalize:
				if is_regression:
					y = y - y.min() / (y.max() - y.min())
				else:
					for j, y_ in enumerate(np.unique(y)):
						y[y == y_] = j

			scores = cross_val_score(model, x, y, iterations, folds,
			                         fit_params, scoring_funcs,
			                         significance_levels, verbose)

			ds_df = pd.DataFrame(scores)
			ds_df['model'] = model.__class__.__name__
			try:
				ds_df['data_set'] = csv_file.split('/')[-1]
			except:
				ds_df['data_set'] = csv_file

			df = df.append(ds_df)

	return df


# -----------------------------------------------------------------------------
# Validity measures
# -----------------------------------------------------------------------------
def reg_n_correct(prediction, y, significance=None):
	"""Calculates the number of correct predictions made by a conformal
	regression model.
	"""
	if significance is not None:
		idx = int(significance * 100 - 1)
		prediction = prediction[:, :, idx]

	low = y >= prediction[:, 0]
	high = y <= prediction[:, 1]
	correct = low * high

	return y[correct].size


def reg_mean_errors(prediction, y, significance):
	"""Calculates the average error rate of a conformal regression model.
	"""
	return 1 - reg_n_correct(prediction, y, significance) / y.size


def class_n_correct(prediction, y, significance):
	"""Calculates the number of correct predictions made by a conformal
	classification model.
	"""
	labels, y = np.unique(y, return_inverse=True)
	prediction = prediction > significance
	correct = np.zeros((y.size,), dtype=bool)
	for i, y_ in enumerate(y):
		correct[i] = prediction[i, int(y_)]
	return np.sum(correct)


def class_mean_errors(prediction, y, significance=None):
	"""Calculates the average error rate of a conformal classification model.
	"""
	return 1 - (class_n_correct(prediction, y, significance) / y.size)


def class_one_err(prediction, y, significance=None):
	"""Calculates the error rate of conformal classifier predictions containing
	 only a single output label.
	"""
	labels, y = np.unique(y, return_inverse=True)
	prediction = prediction > significance
	idx = np.arange(0, y.size, 1)
	idx = filter(lambda x: np.sum(prediction[x, :]) == 1, idx)
	errors = filter(lambda x: not prediction[x, int(y[x])], idx)

	if len(idx) > 0:
		return np.size(errors) / np.size(idx)
	else:
		return 0


def class_mean_errors_one_class(prediction, y, significance, c=0):
	"""Calculates the average error rate of a conformal classification model,
	  considering only test examples belonging to class ``c``. Use
	  ``functools.partial`` in order to test other classes.
	"""
	labels, y = np.unique(y, return_inverse=True)
	prediction = prediction > significance
	idx = np.arange(0, y.size, 1)[y == c]
	errs = np.sum(1 for _ in filter(lambda x: not prediction[x, c], idx))

	if idx.size > 0:
		return errs / idx.size
	else:
		return 0


def class_one_err_one_class(prediction, y, significance, c=0):
	"""Calculates the error rate of conformal classifier predictions containing
	 only a single output label. Considers only test examples belonging to
	 class ``c``. Use ``functools.partial`` in order to test other classes.
	"""
	labels, y = np.unique(y, return_inverse=True)
	prediction = prediction > significance
	idx = np.arange(0, y.size, 1)
	idx = filter(lambda x: prediction[x, c], idx)
	idx = filter(lambda x: np.sum(prediction[x, :]) == 1, idx)
	errors = filter(lambda x: int(y[x]) != c, idx)

	if len(idx) > 0:
		return np.size(errors) / np.size(idx)
	else:
		return 0


# -----------------------------------------------------------------------------
# Efficiency measures
# -----------------------------------------------------------------------------
def _reg_interval_size(prediction, y, significance):
	idx = int(significance * 100 - 1)
	prediction = prediction[:, :, idx]

	return prediction[:, 1] - prediction[:, 0]


def reg_min_size(prediction, y, significance):
	return np.min(_reg_interval_size(prediction, y, significance))


def reg_q1_size(prediction, y, significance):
	return np.percentile(_reg_interval_size(prediction, y, significance), 25)


def reg_median_size(prediction, y, significance):
	return np.median(_reg_interval_size(prediction, y, significance))


def reg_q3_size(prediction, y, significance):
	return np.percentile(_reg_interval_size(prediction, y, significance), 75)


def reg_max_size(prediction, y, significance):
	return np.max(_reg_interval_size(prediction, y, significance))


def reg_mean_size(prediction, y, significance):
	"""Calculates the average prediction interval size of a conformal
	regression model.
	"""
	return np.mean(_reg_interval_size(prediction, y, significance))


def class_avg_c(prediction, y, significance):
	"""Calculates the average number of classes per prediction of a conformal
	classification model.
	"""
	prediction = prediction > significance
	return np.sum(prediction) / prediction.shape[0]


def class_mean_p_val(prediction, y, significance):
	"""Calculates the mean of the p-values output by a conformal classification
	model.
	"""
	return np.mean(prediction)


def class_one_c(prediction, y, significance):
	"""Calculates the rate of singleton predictions (prediction sets containing
	only a single class label) of a conformal classification model.
	"""
	prediction = prediction > significance
	n_singletons = np.sum(1 for _ in filter(lambda x: np.sum(x) == 1,
	                                        prediction))
	return n_singletons / y.size


def class_empty(prediction, y, significance):
	"""Calculates the rate of singleton predictions (prediction sets containing
	only a single class label) of a conformal classification model.
	"""
	prediction = prediction > significance
	n_empty = np.sum(1 for _ in filter(lambda x: np.sum(x) == 0,
	                                        prediction))
	return n_empty / y.size


def n_test(prediction, y, significance):
	"""Provides the number of test patters used in the evaluation.
	"""
	return y.size


#!/usr/bin/env python

"""
Inductive conformal predictors.
"""

# Authors: Henrik Linusson

from __future__ import division
from collections import defaultdict
from functools import partial

import numpy as np
from sklearn.base import BaseEstimator

from .base import RegressorMixin, ClassifierMixin
from .util import calc_p


# -----------------------------------------------------------------------------
# Base inductive conformal predictor
# -----------------------------------------------------------------------------
class BaseIcp(BaseEstimator):
	"""Base class for inductive conformal predictors.
	"""
	def __init__(self, nc_function, condition=None):
		self.cal_x, self.cal_y = None, None
		self.nc_function = nc_function

		# Check if condition-parameter is the default function (i.e.,
		# lambda x: 0). This is so we can safely clone the object without
		# the clone accidentally having self.conditional = True.
		default_condition = lambda x: 0
		is_default = (callable(condition) and
		              (condition.__code__.co_code ==
		               default_condition.__code__.co_code))

		if is_default:
			self.condition = condition
			self.conditional = False
		elif callable(condition):
			self.condition = condition
			self.conditional = True
		else:
			self.condition = lambda x: 0
			self.conditional = False

	def fit(self, x, y):
		"""Fit underlying nonconformity scorer.

		Parameters
		----------
		x : numpy array of shape [n_samples, n_features]
			Inputs of examples for fitting the nonconformity scorer.

		y : numpy array of shape [n_samples]
			Outputs of examples for fitting the nonconformity scorer.

		Returns
		-------
		None
		"""
		# TODO: incremental?
		self.nc_function.fit(x, y)

	def calibrate(self, x, y, increment=False):
		"""Calibrate conformal predictor based on underlying nonconformity
		scorer.

		Parameters
		----------
		x : numpy array of shape [n_samples, n_features]
			Inputs of examples for calibrating the conformal predictor.

		y : numpy array of shape [n_samples, n_features]
			Outputs of examples for calibrating the conformal predictor.

		increment : boolean
			If ``True``, performs an incremental recalibration of the conformal
			predictor. The supplied ``x`` and ``y`` are added to the set of
			previously existing calibration examples, and the conformal
			predictor is then calibrated on both the old and new calibration
			examples.

		Returns
		-------
		None
		"""
		self._calibrate_hook(x, y, increment)
		self._update_calibration_set(x, y, increment)

		if self.conditional:
			category_map = np.array([self.condition((x[i, :], y[i]))
									 for i in range(y.size)])
			self.categories = np.unique(category_map)
			self.cal_scores = defaultdict(partial(np.ndarray, 0))

			for cond in self.categories:
				idx = category_map == cond
				cal_scores = self.nc_function.score(self.cal_x[idx, :],
				                                    self.cal_y[idx])
				self.cal_scores[cond] = np.sort(cal_scores,0)[::-1]
		else:
			self.categories = np.array([0])
			cal_scores = self.nc_function.score(self.cal_x, self.cal_y)
			self.cal_scores = {0: np.sort(cal_scores,0)[::-1]}

	def _calibrate_hook(self, x, y, increment):
		pass

	def _update_calibration_set(self, x, y, increment):
		if increment and self.cal_x is not None and self.cal_y is not None:
			self.cal_x = np.vstack([self.cal_x, x])
			self.cal_y = np.hstack([self.cal_y, y])
		else:
			self.cal_x, self.cal_y = x, y


# -----------------------------------------------------------------------------
# Inductive conformal classifier
# -----------------------------------------------------------------------------
class IcpClassifier(BaseIcp, ClassifierMixin):
	"""Inductive conformal classifier.

	Parameters
	----------
	nc_function : BaseScorer
		Nonconformity scorer object used to calculate nonconformity of
		calibration examples and test patterns. Should implement ``fit(x, y)``
		and ``calc_nc(x, y)``.

	smoothing : boolean
		Decides whether to use stochastic smoothing of p-values.

	Attributes
	----------
	cal_x : numpy array of shape [n_cal_examples, n_features]
		Inputs of calibration set.

	cal_y : numpy array of shape [n_cal_examples]
		Outputs of calibration set.

	nc_function : BaseScorer
		Nonconformity scorer object used to calculate nonconformity scores.

	classes : numpy array of shape [n_classes]
		List of class labels, with indices corresponding to output columns
		 of IcpClassifier.predict()

	See also
	--------
	IcpRegressor

	References
	----------
	.. [1] Papadopoulos, H., & Haralambous, H. (2011). Reliable prediction
		intervals with regression neural networks. Neural Networks, 24(8),
		842-851.

	Examples
	--------
	>>> import numpy as np
	>>> from sklearn.datasets import load_iris
	>>> from sklearn.tree import DecisionTreeClassifier
	>>> from nonconformist.base import ClassifierAdapter
	>>> from nonconformist.icp import IcpClassifier
	>>> from nonconformist.nc import ClassifierNc, MarginErrFunc
	>>> iris = load_iris()
	>>> idx = np.random.permutation(iris.target.size)
	>>> train = idx[:int(idx.size / 3)]
	>>> cal = idx[int(idx.size / 3):int(2 * idx.size / 3)]
	>>> test = idx[int(2 * idx.size / 3):]
	>>> model = ClassifierAdapter(DecisionTreeClassifier())
	>>> nc = ClassifierNc(model, MarginErrFunc())
	>>> icp = IcpClassifier(nc)
	>>> icp.fit(iris.data[train, :], iris.target[train])
	>>> icp.calibrate(iris.data[cal, :], iris.target[cal])
	>>> icp.predict(iris.data[test, :], significance=0.10)
	...             # doctest: +SKIP
	array([[ True, False, False],
		[False,  True, False],
		...,
		[False,  True, False],
		[False,  True, False]], dtype=bool)
	"""
	def __init__(self, nc_function, condition=None, smoothing=True):
		super(IcpClassifier, self).__init__(nc_function, condition)
		self.classes = None
		self.smoothing = smoothing

	def _calibrate_hook(self, x, y, increment=False):
		self._update_classes(y, increment)

	def _update_classes(self, y, increment):
		if self.classes is None or not increment:
			self.classes = np.unique(y)
		else:
			self.classes = np.unique(np.hstack([self.classes, y]))

	def predict(self, x, significance=None):
		"""Predict the output values for a set of input patterns.

		Parameters
		----------
		x : numpy array of shape [n_samples, n_features]
			Inputs of patters for which to predict output values.

		significance : float or None
			Significance level (maximum allowed error rate) of predictions.
			Should be a float between 0 and 1. If ``None``, then the p-values
			are output rather than the predictions.

		Returns
		-------
		p : numpy array of shape [n_samples, n_classes]
			If significance is ``None``, then p contains the p-values for each
			sample-class pair; if significance is a float between 0 and 1, then
			p is a boolean array denoting which labels are included in the
			prediction sets.
		"""
		# TODO: if x == self.last_x ...
		n_test_objects = x.shape[0]
		p = np.zeros((n_test_objects, self.classes.size))

		ncal_ngt_neq = self._get_stats(x)

		for i in range(len(self.classes)):
			for j in range(n_test_objects):
				p[j, i] = calc_p(ncal_ngt_neq[j, i, 0],
				                 ncal_ngt_neq[j, i, 1],
				                 ncal_ngt_neq[j, i, 2],
				                 self.smoothing)

		if significance is not None:
			return p > significance
		else:
			return p

	def _get_stats(self, x):
		n_test_objects = x.shape[0]
		ncal_ngt_neq = np.zeros((n_test_objects, self.classes.size, 3))
		for i, c in enumerate(self.classes):
			test_class = np.zeros(x.shape[0], dtype=self.classes.dtype)
			test_class.fill(c)

			# TODO: maybe calculate p-values using cython or similar
			# TODO: interpolated p-values

			# TODO: nc_function.calc_nc should take X * {y1, y2, ... ,yn}
			test_nc_scores = self.nc_function.score(x, test_class)
			for j, nc in enumerate(test_nc_scores):
				cal_scores = self.cal_scores[self.condition((x[j, :], c))][::-1]
				n_cal = cal_scores.size

				idx_left = np.searchsorted(cal_scores, nc, 'left')
				idx_right = np.searchsorted(cal_scores, nc, 'right')

				ncal_ngt_neq[j, i, 0] = n_cal
				ncal_ngt_neq[j, i, 1] = n_cal - idx_right
				ncal_ngt_neq[j, i, 2] = idx_right - idx_left

		return ncal_ngt_neq

	def predict_conf(self, x):
		"""Predict the output values for a set of input patterns, using
		the confidence-and-credibility output scheme.

		Parameters
		----------
		x : numpy array of shape [n_samples, n_features]
			Inputs of patters for which to predict output values.

		Returns
		-------
		p : numpy array of shape [n_samples, 3]
			p contains three columns: the first column contains the most
			likely class for each test pattern; the second column contains
			the confidence in the predicted class label, and the third column
			contains the credibility of the prediction.
		"""
		p = self.predict(x, significance=None)
		label = p.argmax(axis=1)
		credibility = p.max(axis=1)
		for i, idx in enumerate(label):
			p[i, idx] = -np.inf
		confidence = 1 - p.max(axis=1)

		return np.array([label, confidence, credibility]).T


# -----------------------------------------------------------------------------
# Inductive conformal regressor
# -----------------------------------------------------------------------------
class IcpRegressor(BaseIcp, RegressorMixin):
	"""Inductive conformal regressor.

	Parameters
	----------
	nc_function : BaseScorer
		Nonconformity scorer object used to calculate nonconformity of
		calibration examples and test patterns. Should implement ``fit(x, y)``,
		``calc_nc(x, y)`` and ``predict(x, nc_scores, significance)``.

	Attributes
	----------
	cal_x : numpy array of shape [n_cal_examples, n_features]
		Inputs of calibration set.

	cal_y : numpy array of shape [n_cal_examples]
		Outputs of calibration set.

	nc_function : BaseScorer
		Nonconformity scorer object used to calculate nonconformity scores.

	See also
	--------
	IcpClassifier

	References
	----------
	.. [1] Papadopoulos, H., Proedrou, K., Vovk, V., & Gammerman, A. (2002).
		Inductive confidence machines for regression. In Machine Learning: ECML
		2002 (pp. 345-356). Springer Berlin Heidelberg.

	.. [2] Papadopoulos, H., & Haralambous, H. (2011). Reliable prediction
		intervals with regression neural networks. Neural Networks, 24(8),
		842-851.

	Examples
	--------
	>>> import numpy as np
	>>> from sklearn.datasets import load_boston
	>>> from sklearn.tree import DecisionTreeRegressor
	>>> from nonconformist.base import RegressorAdapter
	>>> from nonconformist.icp import IcpRegressor
	>>> from nonconformist.nc import RegressorNc, AbsErrorErrFunc
	>>> boston = load_boston()
	>>> idx = np.random.permutation(boston.target.size)
	>>> train = idx[:int(idx.size / 3)]
	>>> cal = idx[int(idx.size / 3):int(2 * idx.size / 3)]
	>>> test = idx[int(2 * idx.size / 3):]
	>>> model = RegressorAdapter(DecisionTreeRegressor())
	>>> nc = RegressorNc(model, AbsErrorErrFunc())
	>>> icp = IcpRegressor(nc)
	>>> icp.fit(boston.data[train, :], boston.target[train])
	>>> icp.calibrate(boston.data[cal, :], boston.target[cal])
	>>> icp.predict(boston.data[test, :], significance=0.10)
	...     # doctest: +SKIP
	array([[  5. ,  20.6],
		[ 15.5,  31.1],
		...,
		[ 14.2,  29.8],
		[ 11.6,  27.2]])
	"""
	def __init__(self, nc_function, condition=None):
		super(IcpRegressor, self).__init__(nc_function, condition)

	def predict(self, x, significance=None):
		"""Predict the output values for a set of input patterns.

		Parameters
		----------
		x : numpy array of shape [n_samples, n_features]
			Inputs of patters for which to predict output values.

		significance : float
			Significance level (maximum allowed error rate) of predictions.
			Should be a float between 0 and 1. If ``None``, then intervals for
			all significance levels (0.01, 0.02, ..., 0.99) are output in a
			3d-matrix.

		Returns
		-------
		p : numpy array of shape [n_samples, 2] or [n_samples, 2, 99}
			If significance is ``None``, then p contains the interval (minimum
			and maximum boundaries) for each test pattern, and each significance
			level (0.01, 0.02, ..., 0.99). If significance is a float between
			0 and 1, then p contains the prediction intervals (minimum and
			maximum	boundaries) for the set of test patterns at the chosen
			significance level.
		"""
		# TODO: interpolated p-values

		n_significance = (99 if significance is None
		                  else np.array(significance).size)

		if n_significance > 1:
			prediction = np.zeros((x.shape[0], 2, n_significance))
		else:
			prediction = np.zeros((x.shape[0], 2))

		condition_map = np.array([self.condition((x[i, :], None))
		                          for i in range(x.shape[0])])

		for condition in self.categories:
			idx = condition_map == condition
			if np.sum(idx) > 0:
				p = self.nc_function.predict(x[idx, :],
				                             self.cal_scores[condition],
				                             significance)
				if n_significance > 1:
					prediction[idx, :, :] = p
				else:
					prediction[idx, :] = p

		return prediction


class OobCpClassifier(IcpClassifier):
	def __init__(self,
	             nc_function,
	             condition=None,
	             smoothing=True):
		super(OobCpClassifier, self).__init__(nc_function,
		                                      condition,
		                                      smoothing)

	def fit(self, x, y):
		super(OobCpClassifier, self).fit(x, y)
		super(OobCpClassifier, self).calibrate(x, y, False)

	def calibrate(self, x, y, increment=False):
		# Should throw exception (or really not be implemented for oob)
		pass


class OobCpRegressor(IcpRegressor):
	def __init__(self,
				 nc_function,
				 condition=None):
		super(OobCpRegressor, self).__init__(nc_function,
											  condition)

	def fit(self, x, y):
		super(OobCpRegressor, self).fit(x, y)
		super(OobCpRegressor, self).calibrate(x, y, False)

	def calibrate(self, x, y, increment=False):
		# Should throw exception (or really not be implemented for oob)
		pass



from .base.base import Base
from .base.baseRegressor import BaseRegressor
from .boosting.adaBoostClassifier import AdaBoostClassifier
from .custom.customClassifier import CustomClassifier
from .custom.customRegressor import CustomRegressor
from .datasets import Downloader
from .deep.deepClassifier import DeepClassifier
from .deep.deepRegressor import DeepRegressor
from .deep.deepMTS import DeepMTS
from .glm.glmClassifier import GLMClassifier
from .glm.glmRegressor import GLMRegressor
from .lazypredict.lazyClassifier import LazyClassifier
from .lazypredict.lazyRegressor import LazyRegressor
from .lazypredict.lazyMTS import LazyMTS
from .lazypredict.lazydeepClassifier import LazyDeepClassifier
from .lazypredict.lazydeepRegressor import LazyDeepRegressor
from .lazypredict.lazydeepMTS import LazyDeepMTS
from .mts.mts import MTS
from .multitask.multitaskClassifier import MultitaskClassifier
from .multitask.simplemultitaskClassifier import SimpleMultitaskClassifier
from .optimizers.optimizer import Optimizer
from .predictioninterval import PredictionInterval
from .randombag.randomBagClassifier import RandomBagClassifier
from .randombag.randomBagRegressor import RandomBagRegressor
from .ridge2.ridge2Classifier import Ridge2Classifier
from .ridge2.ridge2Regressor import Ridge2Regressor
from .ridge2.ridge2MultitaskClassifier import Ridge2MultitaskClassifier
from .rvfl.bayesianrvflRegressor import BayesianRVFLRegressor
from .rvfl.bayesianrvfl2Regressor import BayesianRVFL2Regressor
from .sampling import SubSampler

__all__ = [
    "AdaBoostClassifier",
    "Base",
    "BaseRegressor",
    "BayesianRVFLRegressor",
    "BayesianRVFL2Regressor",
    "CustomClassifier",
    "CustomRegressor",
    "DeepClassifier",
    "DeepRegressor",
    "DeepMTS",
    "Downloader",
    "GLMClassifier",
    "GLMRegressor",
    "LazyClassifier",
    "LazyRegressor",
    "LazyMTS",
    "LazyDeepClassifier",
    "LazyDeepRegressor",
    "LazyDeepMTS",
    "MTS",
    "MultitaskClassifier",
    "PredictionInterval",
    "SimpleMultitaskClassifier",
    "Optimizer",
    "RandomBagRegressor",
    "RandomBagClassifier",
    "Ridge2Regressor",
    "Ridge2Classifier",
    "Ridge2MultitaskClassifier",
    "SubSampler",
]



from .ridge2Classifier import Ridge2Classifier
from .ridge2Regressor import Ridge2Regressor
from .ridge2MultitaskClassifier import Ridge2MultitaskClassifier

__all__ = ["Ridge2Classifier", "Ridge2Regressor", "Ridge2MultitaskClassifier"]



from __future__ import division
import numpy as np

def calc_p(ncal, ngt, neq, smoothing=False):
	if smoothing:
		return (ngt + (neq + 1) * np.random.uniform(0, 1)) / (ncal + 1)
	else:
		return (ngt + neq + 1) / (ncal + 1)



#!/usr/bin/env python

"""
Aggregated conformal predictors
"""

# Authors: Henrik Linusson

import numpy as np
from sklearn.model_selection import KFold, StratifiedKFold
from sklearn.model_selection import ShuffleSplit, StratifiedShuffleSplit
from sklearn.base import clone
from nonconformist.base import BaseEstimator
from nonconformist.util import calc_p


# -----------------------------------------------------------------------------
# Sampling strategies
# -----------------------------------------------------------------------------
class BootstrapSampler(object):
	"""Bootstrap sampler.

	See also
	--------
	CrossSampler, RandomSubSampler

	Examples
	--------
	"""
	def gen_samples(self, y, n_samples, problem_type):
		for i in range(n_samples):
			idx = np.array(range(y.size))
			train = np.random.choice(y.size, y.size, replace=True)
			cal_mask = np.array(np.ones(idx.size), dtype=bool)
			for j in train:
				cal_mask[j] = False
			cal = idx[cal_mask]

			yield train, cal


class CrossSampler(object):
	"""Cross-fold sampler.

	See also
	--------
	BootstrapSampler, RandomSubSampler

	Examples
	--------
	"""
	def gen_samples(self, y, n_samples, problem_type):
		if problem_type == 'classification':
			folds = StratifiedKFold(y, n_folds=n_samples)
		else:
			folds = KFold(y.size, n_folds=n_samples)
		for train, cal in folds:
			yield train, cal


class RandomSubSampler(object):
	"""Random subsample sampler.

	Parameters
	----------
	calibration_portion : float
		Ratio (0-1) of examples to use for calibration.

	See also
	--------
	BootstrapSampler, CrossSampler

	Examples
	--------
	"""
	def __init__(self, calibration_portion=0.3):
		self.cal_portion = calibration_portion

	def gen_samples(self, y, n_samples, problem_type):
		if problem_type == 'classification':
			splits = StratifiedShuffleSplit(y,
			                                n_iter=n_samples,
			                                test_size=self.cal_portion)
		else:
			splits = ShuffleSplit(y.size,
			                      n_iter=n_samples,
			                      test_size=self.cal_portion)

		for train, cal in splits:
			yield train, cal


# -----------------------------------------------------------------------------
# Conformal ensemble
# -----------------------------------------------------------------------------
class AggregatedCp(BaseEstimator):
	"""Aggregated conformal predictor.

	Combines multiple IcpClassifier or IcpRegressor predictors into an
	aggregated model.

	Parameters
	----------
	predictor : object
		Prototype conformal predictor (e.g. IcpClassifier or IcpRegressor)
		used for defining conformal predictors included in the aggregate model.

	sampler : object
		Sampler object used to generate training and calibration examples
		for the underlying conformal predictors.

	aggregation_func : callable
		Function used to aggregate the predictions of the underlying
		conformal predictors. Defaults to ``numpy.mean``.

	n_models : int
		Number of models to aggregate.

	Attributes
	----------
	predictor : object
		Prototype conformal predictor.

	predictors : list
		List of underlying conformal predictors.

	sampler : object
		Sampler object used to generate training and calibration examples.

	agg_func : callable
		Function used to aggregate the predictions of the underlying
		conformal predictors

	References
	----------
	.. [1] Vovk, V. (2013). Cross-conformal predictors. Annals of Mathematics
		and Artificial Intelligence, 1-20.

	.. [2] Carlsson, L., Eklund, M., & Norinder, U. (2014). Aggregated
		Conformal Prediction. In Artificial Intelligence Applications and
		Innovations (pp. 231-240). Springer Berlin Heidelberg.

	Examples
	--------
	"""
	def __init__(self,
	             predictor,
	             sampler=BootstrapSampler(),
	             aggregation_func=None,
	             n_models=10):
		self.predictors = []
		self.n_models = n_models
		self.predictor = predictor
		self.sampler = sampler

		if aggregation_func is not None:
			self.agg_func = aggregation_func
		else:
			self.agg_func = lambda x: np.mean(x, axis=2)

	def fit(self, x, y):
		"""Fit underlying conformal predictors.

		Parameters
		----------
		x : numpy array of shape [n_samples, n_features]
			Inputs of examples for fitting the underlying conformal predictors.

		y : numpy array of shape [n_samples]
			Outputs of examples for fitting the underlying conformal predictors.

		Returns
		-------
		None
		"""
		self.n_train = y.size
		self.predictors = []
		idx = np.random.permutation(y.size)
		x, y = x[idx, :], y[idx]
		problem_type = self.predictor.__class__.get_problem_type()
		samples = self.sampler.gen_samples(y,
		                                   self.n_models,
		                                   problem_type)
		for train, cal in samples:
			predictor = clone(self.predictor)
			predictor.fit(x[train, :], y[train])
			predictor.calibrate(x[cal, :], y[cal])
			self.predictors.append(predictor)

		if problem_type == 'classification':
			self.classes = self.predictors[0].classes

	def predict(self, x, significance=None):
		"""Predict the output values for a set of input patterns.

		Parameters
		----------
		x : numpy array of shape [n_samples, n_features]
			Inputs of patters for which to predict output values.

		significance : float or None
			Significance level (maximum allowed error rate) of predictions.
			Should be a float between 0 and 1. If ``None``, then the p-values
			are output rather than the predictions. Note: ``significance=None``
			is applicable to classification problems only.

		Returns
		-------
		p : numpy array of shape [n_samples, n_classes] or [n_samples, 2]
			For classification problems: If significance is ``None``, then p
			contains the p-values for each sample-class pair; if significance
			is a float between 0 and 1, then p is a boolean array denoting
			which labels are included in the prediction sets.

			For regression problems: Prediction interval (minimum and maximum
			boundaries) for the set of test patterns.
		"""
		is_regression =\
			self.predictor.__class__.get_problem_type() == 'regression'

		n_examples = x.shape[0]

		if is_regression and significance is None:
			signs = np.arange(0.01, 1.0, 0.01)
			pred = np.zeros((n_examples, 2, signs.size))
			for i, s in enumerate(signs):
				predictions = np.dstack([p.predict(x, s)
				                         for p in self.predictors])
				predictions = self.agg_func(predictions)
				pred[:, :, i] = predictions
			return pred
		else:
			def f(p, x):
				return p.predict(x, significance if is_regression else None)
			predictions = np.dstack([f(p, x) for p in self.predictors])
			predictions = self.agg_func(predictions)

			if significance and not is_regression:
				return predictions >= significance
			else:
				return predictions


class CrossConformalClassifier(AggregatedCp):
	"""Cross-conformal classifier.

	Combines multiple IcpClassifiers into a cross-conformal classifier.

	Parameters
	----------
	predictor : object
		Prototype conformal predictor (e.g. IcpClassifier or IcpRegressor)
		used for defining conformal predictors included in the aggregate model.

	aggregation_func : callable
		Function used to aggregate the predictions of the underlying
		conformal predictors. Defaults to ``numpy.mean``.

	n_models : int
		Number of models to aggregate.

	Attributes
	----------
	predictor : object
		Prototype conformal predictor.

	predictors : list
		List of underlying conformal predictors.

	sampler : object
		Sampler object used to generate training and calibration examples.

	agg_func : callable
		Function used to aggregate the predictions of the underlying
		conformal predictors

	References
	----------
	.. [1] Vovk, V. (2013). Cross-conformal predictors. Annals of Mathematics
		and Artificial Intelligence, 1-20.

	Examples
	--------
	"""
	def __init__(self,
				 predictor,
				 n_models=10):
		super(CrossConformalClassifier, self).__init__(predictor,
													   CrossSampler(),
													   n_models)

	def predict(self, x, significance=None):
		ncal_ngt_neq = np.stack([p._get_stats(x) for p in self.predictors],
		                        axis=3)
		ncal_ngt_neq = ncal_ngt_neq.sum(axis=3)

		p = calc_p(ncal_ngt_neq[:, :, 0],
		           ncal_ngt_neq[:, :, 1],
		           ncal_ngt_neq[:, :, 2],
		           smoothing=self.predictors[0].smoothing)

		if significance:
			return p > significance
		else:
			return p


class BootstrapConformalClassifier(AggregatedCp):
	"""Bootstrap conformal classifier.

	Combines multiple IcpClassifiers into a bootstrap conformal classifier.

	Parameters
	----------
	predictor : object
		Prototype conformal predictor (e.g. IcpClassifier or IcpRegressor)
		used for defining conformal predictors included in the aggregate model.

	aggregation_func : callable
		Function used to aggregate the predictions of the underlying
		conformal predictors. Defaults to ``numpy.mean``.

	n_models : int
		Number of models to aggregate.

	Attributes
	----------
	predictor : object
		Prototype conformal predictor.

	predictors : list
		List of underlying conformal predictors.

	sampler : object
		Sampler object used to generate training and calibration examples.

	agg_func : callable
		Function used to aggregate the predictions of the underlying
		conformal predictors

	References
	----------
	.. [1] Vovk, V. (2013). Cross-conformal predictors. Annals of Mathematics
		and Artificial Intelligence, 1-20.

	Examples
	--------
	"""
	def __init__(self,
				 predictor,
	             n_models=10):
		super(BootstrapConformalClassifier, self).__init__(predictor,
														   BootstrapSampler(),
														   n_models)

	def predict(self, x, significance=None):
		ncal_ngt_neq = np.stack([p._get_stats(x) for p in self.predictors],
		                        axis=3)
		ncal_ngt_neq = ncal_ngt_neq.sum(axis=3)

		p = calc_p(ncal_ngt_neq[:, :, 0] + ncal_ngt_neq[:, :, 0] / self.n_train,
		           ncal_ngt_neq[:, :, 1] + ncal_ngt_neq[:, :, 0] / self.n_train,
		           ncal_ngt_neq[:, :, 2],
		           smoothing=self.predictors[0].smoothing)

		if significance:
			return p > significance
		else:
			return p



#!/usr/bin/env python

"""
docstring
"""

# Authors: Henrik Linusson

import abc
import numpy as np

from sklearn.base import BaseEstimator


class RegressorMixin(object):
	def __init__(self):
		super(RegressorMixin, self).__init__()

	@classmethod
	def get_problem_type(cls):
		return 'regression'


class ClassifierMixin(object):
	def __init__(self):
		super(ClassifierMixin, self).__init__()

	@classmethod
	def get_problem_type(cls):
		return 'classification'


class BaseModelAdapter(BaseEstimator):
	__metaclass__ = abc.ABCMeta

	def __init__(self, model, fit_params=None):
		super(BaseModelAdapter, self).__init__()

		self.model = model
		self.last_x, self.last_y = None, None
		self.clean = False
		self.fit_params = {} if fit_params is None else fit_params

	def fit(self, x, y):
		"""Fits the model.

		Parameters
		----------
		x : numpy array of shape [n_samples, n_features]
			Inputs of examples for fitting the model.

		y : numpy array of shape [n_samples]
			Outputs of examples for fitting the model.

		Returns
		-------
		None
		"""

		self.model.fit(x, y, **self.fit_params)
		self.clean = False

	def predict(self, x):
		"""Returns the prediction made by the underlying model.

		Parameters
		----------
		x : numpy array of shape [n_samples, n_features]
			Inputs of test examples.

		Returns
		-------
		y : numpy array of shape [n_samples]
			Predicted outputs of test examples.
		"""
		if (
			not self.clean or
			self.last_x is None or
			self.last_y is None or
			not np.array_equal(self.last_x, x)
		):
			self.last_x = x
			self.last_y = self._underlying_predict(x)
			self.clean = True

		return self.last_y.copy()

	@abc.abstractmethod
	def _underlying_predict(self, x):
		"""Produces a prediction using the encapsulated model.

		Parameters
		----------
		x : numpy array of shape [n_samples, n_features]
			Inputs of test examples.

		Returns
		-------
		y : numpy array of shape [n_samples]
			Predicted outputs of test examples.
		"""
		pass


class ClassifierAdapter(BaseModelAdapter):
	def __init__(self, model, fit_params=None):
		super(ClassifierAdapter, self).__init__(model, fit_params)

	def _underlying_predict(self, x):
		return self.model.predict_proba(x)


class RegressorAdapter(BaseModelAdapter):
	def __init__(self, model, fit_params=None):
		super(RegressorAdapter, self).__init__(model, fit_params)

	def _underlying_predict(self, x):
		return self.model.predict(x)


class OobMixin(object):
	def __init__(self, model, fit_params=None):
		super(OobMixin, self).__init__(model, fit_params)
		self.train_x = None

	def fit(self, x, y):
		super(OobMixin, self).fit(x, y)
		self.train_x = x

	def _underlying_predict(self, x):
		# TODO: sub-sampling of ensemble for test patterns
		oob = x == self.train_x

		if hasattr(oob, 'all'):
			oob = oob.all()

		if oob:
			return self._oob_prediction()
		else:
			return super(OobMixin, self)._underlying_predict(x)


class OobClassifierAdapter(OobMixin, ClassifierAdapter):
	def __init__(self, model, fit_params=None):
		super(OobClassifierAdapter, self).__init__(model, fit_params)

	def _oob_prediction(self):
		return self.model.oob_decision_function_


class OobRegressorAdapter(OobMixin, RegressorAdapter):
	def __init__(self, model, fit_params=None):
		super(OobRegressorAdapter, self).__init__(model, fit_params)

	def _oob_prediction(self):
		return self.model.oob_prediction_



# Authors: T. Moudiki
#
# License: BSD 3 Clause Clear

from .helpers import scd, sgd, one_hot_encode


class Optimizer:
    """Optimizer class

    Attributes:

        type_optim: str
            type of optimizer, (currently) either 'sgd' (stochastic minibatch gradient descent)
            or 'scd' (stochastic minibatch coordinate descent)

        num_iters: int
            number of iterations of the optimizer

        learning_rate: float
            step size

        batch_prop: float
            proportion of the initial data used at each optimization step

        learning_method: str
            "poly" - learning rate decreasing as a polynomial function
            of # of iterations (default)
            "exp" - learning rate decreasing as an exponential function
            of # of iterations
            "momentum" - gradient descent using momentum

        randomization: str
            type of randomization applied at each step
            "strat" - stratified subsampling (default)
            "shuffle" - random subsampling

        mass: float
            mass on velocity, for `method` == "momentum"

        decay: float
            coefficient of decrease of the learning rate for
            `method` == "poly" and `method` == "exp"

        tolerance: float
            early stopping parameter (convergence of loss function)

        verbose: int
            controls verbosity of gradient descent
            0 - nothing is printed
            1 - a progress bar is printed
            2 - successive loss function values are printed

    """

    # construct the object -----

    def __init__(
        self,
        type_optim="sgd",
        num_iters=100,
        learning_rate=0.01,
        batch_prop=1.0,
        learning_method="momentum",
        randomization="strat",
        mass=0.9,
        decay=0.1,
        tolerance=1e-3,
        verbose=1,
    ):
        self.type_optim = type_optim
        self.num_iters = num_iters
        self.learning_rate = learning_rate
        self.batch_prop = batch_prop
        self.learning_method = learning_method
        self.randomization = randomization
        self.mass = mass
        self.decay = decay
        self.tolerance = tolerance
        self.verbose = verbose
        self.opt = None

    def fit(self, loss_func, response, x0, **kwargs):
        """Fit GLM model to training data (X, y).

        Args:

            loss_func: loss function

            response: array-like, shape = [n_samples]
            target variable (used for subsampling)

            x0: array-like, shape = [n_features]
                initial value provided to the optimizer

            **kwargs: additional parameters to be passed to
                    loss function

        Returns:

            self: object

        """

        if self.type_optim == "scd":
            self.results = scd(
                loss_func,
                response=response,
                x=x0,
                num_iters=self.num_iters,
                batch_prop=self.batch_prop,
                learning_rate=self.learning_rate,
                learning_method=self.learning_method,
                mass=self.mass,
                decay=self.decay,
                randomization=self.randomization,
                tolerance=self.tolerance,
                verbose=self.verbose,
                **kwargs
            )

        if self.type_optim == "sgd":
            self.results = sgd(
                loss_func,
                response=response,
                x=x0,
                num_iters=self.num_iters,
                batch_prop=self.batch_prop,
                learning_rate=self.learning_rate,
                learning_method=self.learning_method,
                mass=self.mass,
                decay=self.decay,
                randomization=self.randomization,
                tolerance=self.tolerance,
                verbose=self.verbose,
                **kwargs
            )

        return self

    def one_hot_encode(self, y, n_classes):
        return one_hot_encode(y, n_classes)



# Authors: Thierry Moudiki
#
# License: BSD 3

import numpy as np
import sklearn.metrics as skm2
from .custom import Custom
from ..utils import matrixops as mo
from ..predictioninterval import PredictionInterval
from sklearn.base import RegressorMixin
from functools import partial
from scipy.stats import norm


class CustomRegressor(Custom, RegressorMixin):
    """Custom Regression model

    This class is used to 'augment' any regression model with transformed features.

    Parameters:

        obj: object
            any object containing a method fit (obj.fit()) and a method predict
            (obj.predict())

        n_hidden_features: int
            number of nodes in the hidden layer

        activation_name: str
            activation function: 'relu', 'tanh', 'sigmoid', 'prelu' or 'elu'

        a: float
            hyperparameter for 'prelu' or 'elu' activation function

        nodes_sim: str
            type of simulation for the nodes: 'sobol', 'hammersley', 'halton',
            'uniform'

        bias: boolean
            indicates if the hidden layer contains a bias term (True) or not
            (False)

        dropout: float
            regularization parameter; (random) percentage of nodes dropped out
            of the training

        direct_link: boolean
            indicates if the original predictors are included (True) in model's
            fitting or not (False)

        n_clusters: int
            number of clusters for 'kmeans' or 'gmm' clustering (could be 0:
                no clustering)

        cluster_encode: bool
            defines how the variable containing clusters is treated (default is one-hot)
            if `False`, then labels are used, without one-hot encoding

        type_clust: str
            type of clustering method: currently k-means ('kmeans') or Gaussian
            Mixture Model ('gmm')

        type_scaling: a tuple of 3 strings
            scaling methods for inputs, hidden layer, and clustering respectively
            (and when relevant).
            Currently available: standardization ('std') or MinMax scaling ('minmax')
        
        type_pi: str.            
            type of prediction interval; currently "kde" (default) or "bootstrap".
            Used only in `self.predict`, for `self.replications` > 0 and `self.kernel` 
            in ('gaussian', 'tophat'). Default is `None`.
        
        replications: int.
            number of replications (if needed) for predictive simulation. 
            Used only in `self.predict`, for `self.kernel` in ('gaussian', 
            'tophat') and `self.type_pi = 'kde'`. Default is `None`.

        kernel: str.
            the kernel to use for kernel density estimation (used for predictive 
            simulation in `self.predict`, with `method='splitconformal'` and 
            `type_pi = 'kde'`). Currently, either 'gaussian' or 'tophat'.

        col_sample: float
            percentage of covariates randomly chosen for training

        row_sample: float
            percentage of rows chosen for training, by stratified bootstrapping

        seed: int
            reproducibility seed for nodes_sim=='uniform'

        type_fit: str
            'regression'

        backend: str
            "cpu" or "gpu" or "tpu"

    Examples:

    See [https://thierrymoudiki.github.io/blog/2024/03/18/python/conformal-and-bayesian-regression](https://thierrymoudiki.github.io/blog/2024/03/18/python/conformal-and-bayesian-regression)

    """

    # construct the object -----

    def __init__(
        self,
        obj,
        n_hidden_features=5,
        activation_name="relu",
        a=0.01,
        nodes_sim="sobol",
        bias=True,
        dropout=0,
        direct_link=True,
        n_clusters=2,
        cluster_encode=True,
        type_clust="kmeans",
        type_scaling=("std", "std", "std"),
        type_pi=None,  
        replications=None,              
        kernel=None,        
        col_sample=1,
        row_sample=1,
        seed=123,
        backend="cpu",
    ):
        super().__init__(
            obj=obj,
            n_hidden_features=n_hidden_features,
            activation_name=activation_name,
            a=a,
            nodes_sim=nodes_sim,
            bias=bias,
            dropout=dropout,
            direct_link=direct_link,
            n_clusters=n_clusters,
            cluster_encode=cluster_encode,
            type_clust=type_clust,
            type_scaling=type_scaling,
            col_sample=col_sample,
            row_sample=row_sample,
            seed=seed,
            backend=backend,
        )

        self.type_fit = "regression"
        self.type_pi=type_pi
        self.replications=replications
        self.kernel=kernel        

    def fit(self, X, y, sample_weight=None, **kwargs):
        """Fit custom model to training data (X, y).

        Parameters:

            X: {array-like}, shape = [n_samples, n_features]
                Training vectors, where n_samples is the number
                of samples and n_features is the number of features.

            y: array-like, shape = [n_samples]
                Target values.

            **kwargs: additional parameters to be passed to
                self.cook_training_set or self.obj.fit

        Returns:

            self: object

        """

        centered_y, scaled_Z = self.cook_training_set(y=y, X=X, **kwargs)

        # if sample_weights, else: (must use self.row_index)
        if sample_weight is not None:
            self.obj.fit(
                scaled_Z,
                centered_y,
                sample_weight=np.ravel(sample_weight, order="C")[
                    self.index_row
                ],
                **kwargs
            )

            return self

        self.obj.fit(scaled_Z, centered_y, **kwargs)

        self.X_ = X

        self.y_ = y        

        return self

    def predict(self, X, 
                level=95, 
                method=None,                 
                **kwargs):
        """Predict test data X.

        Parameters:

            X: {array-like}, shape = [n_samples, n_features]
                Training vectors, where n_samples is the number
                of samples and n_features is the number of features.
            
            level: int
                Level of confidence (default = 95)
            
            method: str
                `None`, or 'splitconformal', 'localconformal'  
                prediction (if you specify `return_pi = True`)
            
            **kwargs: additional parameters 
                    `return_pi = True` for conformal prediction, 
                    with `method` in ('splitconformal', 'localconformal') 
                    or `return_std = True` for `self.obj` in 
                    (`sklearn.linear_model.BayesianRidge`, 
                    `sklearn.linear_model.ARDRegressor`,  
                    `sklearn.gaussian_process.GaussianProcessRegressor`)`

        Returns:

            model predictions:
                an array if uncertainty quantification is not requested,
                  or a tuple if with prediction intervals and simulations 
                  if `return_std = True` (mean, standard deviation, 
                  lower and upper prediction interval) or `return_pi = True` 
                  ()

        """

        if "return_std" in kwargs:

            alpha = 100 - level
            pi_multiplier = norm.ppf(1 - alpha / 200)

            if len(X.shape) == 1:

                n_features = X.shape[0]
                new_X = mo.rbind(
                    X.reshape(1, n_features),
                    np.ones(n_features).reshape(1, n_features),
                )

                mean_, std_ = self.obj.predict(
                        self.cook_test_set(new_X, **kwargs), return_std=True
                    )[0]
                
                preds = self.y_mean_ + mean_                 
                lower = self.y_mean_ + (mean_ - pi_multiplier*std_)
                upper = self.y_mean_ + (mean_ + pi_multiplier*std_)

                return preds, std_, lower, upper

            # len(X.shape) > 1
            mean_, std_ = self.obj.predict(
                        self.cook_test_set(X, **kwargs), return_std=True
                    )
                
            preds = self.y_mean_ + mean_                 
            lower = self.y_mean_ + (mean_ - pi_multiplier*std_)
            upper = self.y_mean_ + (mean_ + pi_multiplier*std_)

            return preds, std_, lower, upper

        if "return_pi" in kwargs:
            assert method in ('splitconformal', 'localconformal'), \
                "method must be in ('splitconformal', 'localconformal')"
            self.pi = PredictionInterval(obj = self, 
                                         method=method, 
                                         level=level,
                                         type_pi=self.type_pi,
                                         replications=self.replications,   
                                         kernel=self.kernel,
                                         )            
            self.pi.fit(self.X_, self.y_)
            self.X_ = None
            self.y_ = None 
            preds = self.pi.predict(X, return_pi=True)
            return preds

        # "return_std" not in kwargs
        if len(X.shape) == 1:

            n_features = X.shape[0]
            new_X = mo.rbind(
                X.reshape(1, n_features),
                np.ones(n_features).reshape(1, n_features),
            )

            return (
                self.y_mean_
                + self.obj.predict(
                    self.cook_test_set(new_X, **kwargs), **kwargs
                )
            )[0]

        # len(X.shape) > 1
        return self.y_mean_ + self.obj.predict(
            self.cook_test_set(X, **kwargs), **kwargs
        )


# Authors: Thierry Moudiki
#
# License: BSD 3 Clear

import numpy as np
import sklearn.metrics as skm2
from .glm import GLM
from ..utils import matrixops as mo
from ..utils import misc as mx
from sklearn.base import ClassifierMixin
from ..optimizers import Optimizer
from scipy.special import logsumexp, expit, erf


class GLMClassifier(GLM, ClassifierMixin):
    """Generalized 'linear' models using quasi-randomized networks (classification)

    Parameters:

        n_hidden_features: int
            number of nodes in the hidden layer

        lambda1: float
            regularization parameter for GLM coefficients on original features

        alpha1: float
            controls compromize between l1 and l2 norm of GLM coefficients on original features

        lambda2: float
            regularization parameter for GLM coefficients on nonlinear features

        alpha2: float
            controls compromize between l1 and l2 norm of GLM coefficients on nonlinear features

        activation_name: str
            activation function: 'relu', 'tanh', 'sigmoid', 'prelu' or 'elu'

        a: float
            hyperparameter for 'prelu' or 'elu' activation function

        nodes_sim: str
            type of simulation for the nodes: 'sobol', 'hammersley', 'halton',
            'uniform'

        bias: boolean
            indicates if the hidden layer contains a bias term (True) or not
            (False)

        dropout: float
            regularization parameter; (random) percentage of nodes dropped out
            of the training

        direct_link: boolean
            indicates if the original predictors are included (True) in model's
            fitting or not (False)

        n_clusters: int
            number of clusters for 'kmeans' or 'gmm' clustering (could be 0:
                no clustering)

        cluster_encode: bool
            defines how the variable containing clusters is treated (default is one-hot)
            if `False`, then labels are used, without one-hot encoding

        type_clust: str
            type of clustering method: currently k-means ('kmeans') or Gaussian
            Mixture Model ('gmm')

        type_scaling: a tuple of 3 strings
            scaling methods for inputs, hidden layer, and clustering respectively
            (and when relevant).
            Currently available: standardization ('std') or MinMax scaling ('minmax')

        optimizer: object
            optimizer, from class nnetsauce.Optimizer

        seed: int
            reproducibility seed for nodes_sim=='uniform'

    Attributes:

        beta_: vector
            regression coefficients

    Examples:

    See [https://github.com/Techtonique/nnetsauce/blob/master/examples/glm_classification.py](https://github.com/Techtonique/nnetsauce/blob/master/examples/glm_classification.py)

    """

    # construct the object -----

    def __init__(
        self,
        n_hidden_features=5,
        lambda1=0.01,
        alpha1=0.5,
        lambda2=0.01,
        alpha2=0.5,
        family="expit",
        activation_name="relu",
        a=0.01,
        nodes_sim="sobol",
        bias=True,
        dropout=0,
        direct_link=True,
        n_clusters=2,
        cluster_encode=True,
        type_clust="kmeans",
        type_scaling=("std", "std", "std"),
        optimizer=Optimizer(),
        seed=123,
    ):
        super().__init__(
            n_hidden_features=n_hidden_features,
            lambda1=lambda1,
            alpha1=alpha1,
            lambda2=lambda2,
            alpha2=alpha2,
            activation_name=activation_name,
            a=a,
            nodes_sim=nodes_sim,
            bias=bias,
            dropout=dropout,
            direct_link=direct_link,
            n_clusters=n_clusters,
            cluster_encode=cluster_encode,
            type_clust=type_clust,
            type_scaling=type_scaling,
            optimizer=optimizer,
            seed=seed,
        )

        self.family = family

    def logit_loss(self, Y, row_index, XB):
        self.n_classes = Y.shape[1]  # len(np.unique(y))
        # Y = mo.one_hot_encode2(y, self.n_classes)
        # Y = self.optimizer.one_hot_encode(y, self.n_classes)

        # max_double = 709.0 # only if softmax
        # XB[XB > max_double] = max_double
        XB[XB > 709.0] = 709.0

        if row_index is None:
            return -np.mean(np.sum(Y * XB, axis=1) - logsumexp(XB))

        return -np.mean(np.sum(Y[row_index, :] * XB, axis=1) - logsumexp(XB))

    def expit_erf_loss(self, Y, row_index, XB):
        # self.n_classes = len(np.unique(y))
        # Y = mo.one_hot_encode2(y, self.n_classes)
        # Y = self.optimizer.one_hot_encode(y, self.n_classes)
        self.n_classes = Y.shape[1]

        if row_index is None:
            return -np.mean(np.sum(Y * XB, axis=1) - logsumexp(XB))

        return -np.mean(np.sum(Y[row_index, :] * XB, axis=1) - logsumexp(XB))

    def loss_func(
        self,
        beta,
        group_index,
        X,
        Y,
        y,
        row_index=None,
        type_loss="logit",
        **kwargs
    ):
        res = {
            "logit": self.logit_loss,
            "expit": self.expit_erf_loss,
            "erf": self.expit_erf_loss,
        }

        if row_index is None:
            row_index = range(len(y))
            XB = self.compute_XB(
                X,
                beta=np.reshape(beta, (X.shape[1], self.n_classes), order="F"),
            )

            return res[type_loss](Y, row_index, XB) + self.compute_penalty(
                group_index=group_index, beta=beta
            )

        XB = self.compute_XB(
            X,
            beta=np.reshape(beta, (X.shape[1], self.n_classes), order="F"),
            row_index=row_index,
        )

        return res[type_loss](Y, row_index, XB) + self.compute_penalty(
            group_index=group_index, beta=beta
        )

    def fit(self, X, y, **kwargs):
        """Fit GLM model to training data (X, y).

        Args:

            X: {array-like}, shape = [n_samples, n_features]
                Training vectors, where n_samples is the number
                of samples and n_features is the number of features.

            y: array-like, shape = [n_samples]
                Target values.

            **kwargs: additional parameters to be passed to
                    self.cook_training_set or self.obj.fit

        Returns:

            self: object

        """

        assert mx.is_factor(
            y
        ), "y must contain only integers"  # change is_factor and subsampling everywhere

        self.n_classes_ = len(np.unique(y)) # for compatibility with sklearn 

        self.beta_ = None

        n, p = X.shape

        self.group_index = n * X.shape[1]

        self.n_classes = len(np.unique(y))

        output_y, scaled_Z = self.cook_training_set(y=y, X=X, **kwargs)

        # Y = mo.one_hot_encode2(output_y, self.n_classes)
        Y = self.optimizer.one_hot_encode(output_y, self.n_classes)

        # initialization
        beta_ = np.linalg.lstsq(scaled_Z, Y, rcond=None)[0]

        # optimization
        # fit(self, loss_func, response, x0, **kwargs):
        # loss_func(self, beta, group_index, X, y,
        #          row_index=None, type_loss="gaussian",
        #          **kwargs)
        self.optimizer.fit(
            self.loss_func,
            response=y,
            x0=beta_.flatten(order="F"),
            group_index=self.group_index,
            X=scaled_Z,
            Y=Y,
            y=y,
            type_loss=self.family,
        )

        self.beta_ = self.optimizer.results[0]
        self.classes_ = np.unique(y)

        return self

    def predict(self, X, **kwargs):
        """Predict test data X.

        Args:

            X: {array-like}, shape = [n_samples, n_features]
                Training vectors, where n_samples is the number
                of samples and n_features is the number of features.

            **kwargs: additional parameters to be passed to
                    self.cook_test_set

        Returns:

            model predictions: {array-like}

        """

        return np.argmax(self.predict_proba(X, **kwargs), axis=1)

    def predict_proba(self, X, **kwargs):
        """Predict probabilities for test data X.

        Args:

            X: {array-like}, shape = [n_samples, n_features]
                Training vectors, where n_samples is the number
                of samples and n_features is the number of features.

            **kwargs: additional parameters to be passed to
                    self.cook_test_set

        Returns:

            probability estimates for test data: {array-like}

        """
        if len(X.shape) == 1:
            n_features = X.shape[0]
            new_X = mo.rbind(
                X.reshape(1, n_features),
                np.ones(n_features).reshape(1, n_features),
            )

            Z = self.cook_test_set(new_X, **kwargs)

        else:
            Z = self.cook_test_set(X, **kwargs)

        ZB = mo.safe_sparse_dot(
            Z,
            self.beta_.reshape(
                self.n_classes,
                X.shape[1] + self.n_hidden_features + self.n_clusters,
            ).T,
        )

        if self.family == "logit":
            exp_ZB = np.exp(ZB)

            return exp_ZB / exp_ZB.sum(axis=1)[:, None]

        if self.family == "expit":
            exp_ZB = expit(ZB)

            return exp_ZB / exp_ZB.sum(axis=1)[:, None]

        if self.family == "erf":
            exp_ZB = 0.5 * (1 + erf(ZB))

            return exp_ZB / exp_ZB.sum(axis=1)[:, None]


import copy
import numpy as np
import pandas as pd
import platform
from jax import device_put
from scipy import sparse
from sklearn.preprocessing import StandardScaler, MinMaxScaler, RobustScaler
from sklearn.cluster import KMeans
from sklearn.mixture import GaussianMixture

if platform.system() in ("Linux", "Darwin"):
    import jax.numpy as jnp


# column bind
def cbind(x, y, backend="cpu"):

    if isinstance(x, pd.DataFrame) and isinstance(y, pd.DataFrame):

        return pd.concat([x, y], axis=1)

    else:  # x or y are numpy arrays

        if isinstance(x, pd.DataFrame) == False and isinstance(y, pd.DataFrame):

            if len(x.shape) == 1:

                col_names = ["series0"] + y.columns.tolist()

            else:

                col_names = [
                    "series" + str(i) for i in range(x.shape[1])
                ] + y.columns.tolist()

            if backend in ("gpu", "tpu") and (
                platform.system() in ("Linux", "Darwin")
            ):

                res = jnp.column_stack((x.values, y))

                return pd.DataFrame(res, columns=col_names)

            res = np.column_stack((x.values, y))

            return pd.DataFrame(res, columns=col_names)

        if isinstance(x, pd.DataFrame) and isinstance(y, pd.DataFrame) == False:

            if len(y.shape) == 1:

                col_names = x.columns.tolist() + ["series0"]

            else:

                col_names = x.columns.tolist() + [
                    "series" + str(i) for i in range(x.shape[1])
                ]

            if backend in ("gpu", "tpu") and (
                platform.system() in ("Linux", "Darwin")
            ):

                res = jnp.column_stack((x.values, y))

                return pd.DataFrame(res, columns=col_names)

            res = np.column_stack((x.values, y))

            return pd.DataFrame(res, columns=col_names)

        # x and y are numpy arrays
        if backend in ("gpu", "tpu") and (
            platform.system() in ("Linux", "Darwin")
        ):

            return jnp.column_stack((x, y))

        return np.column_stack((x, y))


# center... response
def center_response(y):
    y_mean = np.mean(y)
    return y_mean, (y - y_mean)


# cluster the covariates
def cluster_covariates(X, n_clusters, seed, type_clust="kmeans", **kwargs):
    if isinstance(X, pd.DataFrame):
        X = copy.deepcopy(X.values.astype(float))

    if type_clust == "kmeans":
        kmeans = KMeans(
            n_clusters=n_clusters, random_state=seed, n_init=10, **kwargs
        )
        kmeans.fit(X)

        return kmeans, kmeans.predict(X)

    elif type_clust == "gmm":
        gmm = GaussianMixture(
            n_components=n_clusters, random_state=seed, **kwargs
        )
        gmm.fit(X)

        return gmm, gmm.predict(X)


def convert_df_to_numeric(df):
    """
    Convert all columns of DataFrame to numeric type using astype with loop.

    Parameters:
        df (pd.DataFrame): Input DataFrame with mixed data types.

    Returns:
        pd.DataFrame: DataFrame with all columns converted to numeric type.
    """
    for column in df.columns:
        # Attempt to convert the column to numeric type using astype
        try:
            df[column] = df[column].astype(float)
        except ValueError:
            print(f"Column '{column}' contains non-numeric values.")
    return df


# computes t(x)%*%y
def crossprod(x, y=None, backend="cpu"):
    # assert on dimensions
    sys_platform = platform.system()
    if backend in ("gpu", "tpu") and (sys_platform in ("Linux", "Darwin")):
        x = device_put(x)
        if y is None:
            return jnp.dot(x.T, x).block_until_ready()
        y = device_put(y)
        return jnp.dot(x.T, y).block_until_ready()
    if y is None:
        return np.dot(x.transpose(), x)
    return np.dot(x.transpose(), y)


def delete_last_columns(x, num_columns, inplace=False):
    """
    Delete the last 'num_columns' columns from a DataFrame.

    Parameters:
        x: pandas DataFrame or tuple or DataFrames.
        num_columns (int): Number of columns to delete from the end.
        inplace (bool): Whether to modify the DataFrame in place. Default is False.

    Returns:
        DataFrame: Modified DataFrame if inplace=False, None otherwise.
    """
    if inplace:
        if isinstance(x, pd.DataFrame):
            x.drop(x.columns[-num_columns:], axis=1, inplace=True)
        if isinstance(x[1], pd.DataFrame):
            for i in range(len(x)):
                x[i].drop(x[i].columns[-num_columns:], axis=1, inplace=True)
    else:
        if isinstance(x, pd.DataFrame):
            return x.drop(x.columns[-num_columns:], axis=1)
        if isinstance(x[1], pd.DataFrame):
            modified_dfs = []
            for i in range(len(x)):
                modified_dfs.append(
                    x[i].drop(x[i].columns[-num_columns:], axis=1)
                )
            return tuple(modified_dfs)


# Obtain this for JAX
# Obtain this for JAX
# Obtain this for JAX
# Obtain this for JAX
# Obtain this for JAX
# dropout
def dropout(x, drop_prob=0, seed=123):
    assert 0 <= drop_prob <= 1

    n, p = x.shape

    if drop_prob == 0:
        return x

    if drop_prob == 1:
        return np.zeros_like(x)

    np.random.seed(seed)
    dropped_indices = np.random.rand(n, p) > drop_prob

    return dropped_indices * x / (1 - drop_prob)


# one-hot encoding
def one_hot_encode(x_clusters, n_clusters):
    assert (
        max(x_clusters) <= n_clusters
    ), "you must have max(x_clusters) <= n_clusters"

    n_obs = len(x_clusters)
    res = np.zeros((n_obs, n_clusters))

    for i in range(n_obs):
        res[i, x_clusters[i]] = 1

    return res


# one-hot encoding
def one_hot_encode2(y, n_classes):
    n_obs = len(y)
    res = np.zeros((n_obs, n_classes))

    for i in range(n_obs):
        res[i, y[i]] = 1

    return res


# row bind
def rbind(x, y, backend="cpu"):
    # if len(x.shape) == 1 or len(y.shape) == 1:
    sys_platform = platform.system()
    if backend in ("gpu", "tpu") and (sys_platform in ("Linux", "Darwin")):
        return jnp.row_stack((x, y))
    return np.row_stack((x, y))


# adapted from sklearn.utils.exmath
def safe_sparse_dot(a, b, backend="cpu", dense_output=False):
    """Dot product that handle the sparse matrix case correctly

    Parameters
    ----------
    a : array or sparse matrix
    b : array or sparse matrix
    dense_output : boolean, (default=False)
        When False, ``a`` and ``b`` both being sparse will yield sparse output.
        When True, output will always be a dense array.

    Returns
    -------
    dot_product : array or sparse matrix
        sparse if ``a`` and ``b`` are sparse and ``dense_output=False``.
    """
    sys_platform = platform.system()

    if backend in ("gpu", "tpu") and (sys_platform in ("Linux", "Darwin")):
        # modif when jax.scipy.sparse available
        return jnp.dot(device_put(a), device_put(b)).block_until_ready()

    #    if backend == "cpu":
    if a.ndim > 2 or b.ndim > 2:
        if sparse.issparse(a):
            # sparse is always 2D. Implies b is 3D+
            # [i, j] @ [k, ..., l, m, n] -> [i, k, ..., l, n]
            b_ = np.rollaxis(b, -2)
            b_2d = b_.reshape((b.shape[-2], -1))
            ret = a @ b_2d
            ret = ret.reshape(a.shape[0], *b_.shape[1:])
        elif sparse.issparse(b):
            # sparse is always 2D. Implies a is 3D+
            # [k, ..., l, m] @ [i, j] -> [k, ..., l, j]
            a_2d = a.reshape(-1, a.shape[-1])
            ret = a_2d @ b
            ret = ret.reshape(*a.shape[:-1], b.shape[1])
        else:
            ret = np.dot(a, b)
    else:
        try:
            ret = a @ b
        except:
            ret = np.dot(a, b)

    if (
        sparse.issparse(a)
        and sparse.issparse(b)
        and dense_output
        and hasattr(ret, "toarray")
    ):
        return ret.toarray()

    return ret


# Obtain this for JAX
# Obtain this for JAX
# Obtain this for JAX
# Obtain this for JAX
# Obtain this for JAX
# scale... covariates
def scale_covariates(X, choice="std", training=True, scaler=None):

    if training == True:

        # scaler must be not None
        if choice == "std":
            scaler = StandardScaler(copy=True, with_mean=True, with_std=True)

        elif choice == "minmax":
            scaler = MinMaxScaler()

        else:
            scaler = RobustScaler(
                copy=True, with_centering=True, with_scaling=True
            )

        scaled_X = scaler.fit_transform(X)

        return scaler, scaled_X

    # training == False:
    # scaler must be not None
    return scaler.transform(X)


# from sklearn.utils.exmath
def squared_norm(x, backend="cpu"):
    """Squared Euclidean or Frobenius norm of x.

    Faster than norm(x) ** 2.

    Parameters
    ----------
    x : array_like

    Returns
    -------
    float
        The Euclidean norm when x is a vector, the Frobenius norm when x
        is a matrix (2-d array).
    """
    sys_platform = platform.system()
    if backend in ("gpu", "tpu") and (sys_platform in ("Linux", "Darwin")):
        x = np.ravel(x, order="K")
        x = device_put(x)
        return jnp.dot(x, x).block_until_ready()

    x = np.ravel(x, order="K")
    return np.dot(x, x)


# computes x%*%t(y)
def tcrossprod(x, y=None, backend="cpu"):
    # assert on dimensions
    sys_platform = platform.system()
    if backend in ("gpu", "tpu") and (sys_platform in ("Linux", "Darwin")):
        x = device_put(x)
        if y is None:
            return jnp.dot(x, x.T).block_until_ready()
        y = device_put(y)
        return jnp.dot(x, y.T).block_until_ready()
    if y is None:
        return np.dot(x, x.transpose())
    return np.dot(x, y.transpose())


# convert vector to numpy array
def to_np_array(X):
    return np.array(X.copy(), ndmin=2)



from locale import normalize
import numpy as np
import pickle
from collections import namedtuple
from sklearn.base import BaseEstimator, RegressorMixin
from sklearn.model_selection import train_test_split
from sklearn.ensemble import ExtraTreesRegressor
from sklearn.preprocessing import StandardScaler
from sklearn.neighbors import KernelDensity
from sklearn.model_selection import GridSearchCV
from scipy.stats import gaussian_kde
from tqdm import tqdm
from ..nonconformist import IcpRegressor
from ..nonconformist import RegressorNc 
from ..nonconformist import RegressorNormalizer, AbsErrorErrFunc
from ..utils import Progbar


class PredictionInterval(BaseEstimator, RegressorMixin):
    """Class PredictionInterval: Obtain prediction intervals.
        
    Attributes:
       
        obj: an object;
            fitted object containing methods `fit` and `predict`

        method: a string;
            method for constructing the prediction intervals. 
            Currently "splitconformal" (default) and "localconformal"

        level: a float;                
            Confidence level for prediction intervals. Default is 95, 
            equivalent to a miscoverage error of 5 (%)
        
        replications: an integer;
            Number of replications for simulated conformal (default is `None`)
        
        type_pi: a string;
            type of prediction interval: currently "kde" (default) or "bootstrap"
        
        seed: an integer;
            Reproducibility of fit (there's a random split between fitting and calibration data)
    """

    def __init__(self, obj, 
                 method="splitconformal", 
                 level=95,
                 type_pi = "bootstrap",
                 replications=None,   
                 kernel=None,
                 agg="mean",               
                 seed=123):

        self.obj = obj
        self.method = method
        self.level = level
        self.type_pi = type_pi
        self.replications = replications
        self.kernel = kernel 
        self.agg = agg 
        self.seed = seed
        self.alpha_ = 1 - self.level/100
        self.quantile_ = None
        self.icp_ = None
        self.calibrated_residuals_ = None
        self.scaled_calibrated_residuals_ = None 
        self.calibrated_residuals_scaler_ = None
        self.kde_ = None 


    def fit(self, X, y):
        """Fit the `method` to training data (X, y).           
        
        Args:

            X: array-like, shape = [n_samples, n_features]; 
                Training set vectors, where n_samples is the number 
                of samples and n_features is the number of features.                

            y: array-like, shape = [n_samples, ]; Target values.
                       
        """       

        X_train, X_calibration, y_train, y_calibration = train_test_split(X, y, 
                                                                          test_size=0.5, 
                                                                          random_state=self.seed)

        if self.method == "splitconformal": 

            n_samples_calibration = X_calibration.shape[0]                        
            self.obj.fit(X_train, y_train)
            preds_calibration = self.obj.predict(X_calibration)
            self.calibrated_residuals_ = y_calibration - preds_calibration
            absolute_residuals = np.abs(self.calibrated_residuals_)   
            self.calibrated_residuals_scaler_ = StandardScaler(with_mean=True, with_std=True)
            self.scaled_calibrated_residuals_ = self.calibrated_residuals_scaler_.fit_transform(self.calibrated_residuals_.reshape(-1, 1)).ravel()
            try: 
                # numpy version >= 1.22
                self.quantile_ = np.quantile(a = absolute_residuals, q = self.level/100, method="higher")                                       
            except:
                # numpy version < 1.22
                self.quantile_ = np.quantile(a = absolute_residuals, q = self.level/100, interpolation="higher")           
                

        if self.method == "localconformal":

            mad_estimator = ExtraTreesRegressor()
            normalizer = RegressorNormalizer(self.obj, mad_estimator, AbsErrorErrFunc())
            nc = RegressorNc(self.obj, AbsErrorErrFunc(), normalizer)        
            self.icp_ = IcpRegressor(nc)
            self.icp_.fit(X_train, y_train) 
            self.icp_.calibrate(X_calibration, y_calibration)
            preds_calibration = self.icp_.predict(X_calibration)            
            self.calibrated_residuals_ = y_calibration - preds_calibration
            absolute_residuals = np.abs(self.calibrated_residuals_)   
            self.calibrated_residuals_scaler_ = StandardScaler(with_mean=True, with_std=True)
            self.scaled_calibrated_residuals_ = self.calibrated_residuals_scaler_.fit_transform(self.calibrated_residuals_.reshape(-1, 1)).ravel()

        return self


    def predict(self, X, return_pi=False):
        """Obtain predictions and prediction intervals            

        Args: 

            X: array-like, shape = [n_samples, n_features]; 
                Testing set vectors, where n_samples is the number 
                of samples and n_features is the number of features. 

            return_pi: boolean               
                Whether the prediction interval is returned or not. 
                Default is False, for compatibility with other _estimators_.
                If True, a tuple containing the predictions + lower and upper 
                bounds is returned.

        """                

        pred = self.obj.predict(X)

        if self.method == "splitconformal":

            if self.replications is None: 

                if return_pi:

                    DescribeResult = namedtuple("DescribeResult", ("mean", "lower", "upper"))                    
                    
                    return DescribeResult(pred, pred - self.quantile_, pred + self.quantile_) 

                else: 

                    return pred
            
            else: #  if self.replications is not None

                assert self.type_pi in ("bootstrap", "kde"),\
                "`self.type_pi` must be in ('bootstrap', 'kde')"                                    

                if self.type_pi == "bootstrap":         
                    np.random.seed(self.seed)               
                    self.residuals_sims_ = np.asarray([np.random.choice(a = self.scaled_calibrated_residuals_, 
                                                            size = X.shape[0]) for _ in range(self.replications)]).T
                    self.sims_ = np.asarray([pred + self.calibrated_residuals_scaler_.scale_[0]*self.residuals_sims_[:, i].ravel() for i in range(self.replications)]).T                                        
                elif self.type_pi == "kde":
                    self.kde_ = gaussian_kde(dataset=self.scaled_calibrated_residuals_)
                    self.sims_ = np.asarray([pred + self.calibrated_residuals_scaler_.scale_[0]*self.kde_.resample(size = X.shape[0], seed=self.seed+i).ravel() for i in range(self.replications)]).T                
                
                self.mean_ = np.mean(self.sims_, axis=1)
                self.lower_ = np.quantile(self.sims_, q=self.alpha_ / 200, axis=1)
                self.upper_ = np.quantile(self.sims_, q=1 - self.alpha_ / 200, axis=1) 

                DescribeResult = namedtuple("DescribeResult", ("mean", "sims", "lower", "upper"))                                       

                return DescribeResult(self.mean_, self.sims_, self.lower_, self.upper_) 

        if self.method == "localconformal":

            if self.replications is None: 

                if return_pi:

                    predictions_bounds = self.icp_.predict(X, significance = 1-self.level)
                    DescribeResult = namedtuple("DescribeResult", ("mean", "lower", "upper")) 
                    return DescribeResult(pred, predictions_bounds[:, 0], predictions_bounds[:, 1])                    

                else:
                    
                    return pred
            
            else: #  if self.replications is not None

                assert self.type_pi in ("bootstrap", "kde"),\
                "`self.type_pi` must be in ('bootstrap', 'kde')"                                    

                if self.type_pi == "bootstrap":         
                    np.random.seed(self.seed)               
                    self.residuals_sims_ = np.asarray([np.random.choice(a = self.scaled_calibrated_residuals_, 
                                                            size = X.shape[0]) for _ in range(self.replications)]).T
                    self.sims_ = np.asarray([pred + self.calibrated_residuals_scaler_.scale_[0]*self.residuals_sims_[:, i].ravel() for i in tqdm(range(self.replications))]).T                                        
                elif self.type_pi == "kde":
                    self.kde_ = gaussian_kde(dataset=self.scaled_calibrated_residuals_)
                    self.sims_ = np.asarray([pred + self.calibrated_residuals_scaler_.scale_[0]*self.kde_.resample(size = X.shape[0], seed=self.seed+i).ravel() for i in tqdm(range(self.replications))]).T                
                
                self.mean_ = np.mean(self.sims_, axis=1)
                self.lower_ = np.quantile(self.sims_, q=self.alpha_ / 200, axis=1)
                self.upper_ = np.quantile(self.sims_, q=1 - self.alpha_ / 200, axis=1) 

                DescribeResult = namedtuple("DescribeResult", ("mean", "sims", "lower", "upper"))                                       

                return DescribeResult(self.mean_, self.sims_, self.lower_, self.upper_) 



# Authors: Thierry Moudiki
#
# License: BSD 3 Clear

import numpy as np
import platform
import sys
import sklearn.metrics as skm2
from scipy.optimize import minimize
from .ridge2 import Ridge2
from ..utils import matrixops as mo
from ..utils import misc as mx
from sklearn.base import ClassifierMixin
from scipy.special import logsumexp


class Ridge2Classifier(Ridge2, ClassifierMixin):
    """Multinomial logit classification with 2 regularization parameters

    Parameters:

        n_hidden_features: int
            number of nodes in the hidden layer

        activation_name: str
            activation function: 'relu', 'tanh', 'sigmoid', 'prelu' or 'elu'

        a: float
            hyperparameter for 'prelu' or 'elu' activation function

        nodes_sim: str
            type of simulation for the nodes: 'sobol', 'hammersley', 'halton',
            'uniform'

        bias: boolean
            indicates if the hidden layer contains a bias term (True) or not
            (False)

        dropout: float
            regularization parameter; (random) percentage of nodes dropped out
            of the training

        direct_link: boolean
            indicates if the original predictors are included (True) in model's
            fitting or not (False)

        n_clusters: int
            number of clusters for 'kmeans' or 'gmm' clustering (could be 0:
                no clustering)

        cluster_encode: bool
            defines how the variable containing clusters is treated (default is one-hot)
            if `False`, then labels are used, without one-hot encoding

        type_clust: str
            type of clustering method: currently k-means ('kmeans') or Gaussian
            Mixture Model ('gmm')

        type_scaling: a tuple of 3 strings
            scaling methods for inputs, hidden layer, and clustering respectively
            (and when relevant).
            Currently available: standardization ('std') or MinMax scaling ('minmax')

        lambda1: float
            regularization parameter on direct link

        lambda2: float
            regularization parameter on hidden layer

        seed: int
            reproducibility seed for nodes_sim=='uniform'

        backend: str
            "cpu" or "gpu" or "tpu"

    Attributes:

        beta_: {array-like}
            regression coefficients


    Examples:

    See also [https://github.com/Techtonique/nnetsauce/blob/master/examples/ridge_classification.py](https://github.com/Techtonique/nnetsauce/blob/master/examples/ridge_classification.py)

    ```python
    import nnetsauce as ns
    import numpy as np
    from sklearn.datasets import load_breast_cancer
    from sklearn.model_selection import train_test_split
    from time import time


    breast_cancer = load_breast_cancer()
    X = breast_cancer.data
    y = breast_cancer.target

    # split data into training test and test set
    np.random.seed(123)
    X_train, X_test, y_train, y_test = train_test_split(X, y, test_size=0.2)

    # create the model with nnetsauce
    fit_obj = ns.Ridge2Classifier(lambda1 = 6.90185578e+04,
                                lambda2 = 3.17392781e+02,
                                n_hidden_features=95,
                                n_clusters=2,
                                dropout = 3.62817383e-01,
                                type_clust = "gmm")

    # fit the model on training set
    start = time()
    fit_obj.fit(X_train, y_train)
    print(f"Elapsed {time() - start}")

    # get the accuracy on test set
    start = time()
    print(fit_obj.score(X_test, y_test))
    print(f"Elapsed {time() - start}")

    # get area under the curve on test set (auc)
    print(fit_obj.score(X_test, y_test, scoring="roc_auc"))
    ```


    """

    # construct the object -----

    def __init__(
        self,
        n_hidden_features=5,
        activation_name="relu",
        a=0.01,
        nodes_sim="sobol",
        bias=True,
        dropout=0,
        direct_link=True,
        n_clusters=2,
        cluster_encode=True,
        type_clust="kmeans",
        type_scaling=("std", "std", "std"),
        lambda1=0.1,
        lambda2=0.1,
        seed=123,
        backend="cpu",
    ):
        super().__init__(
            n_hidden_features=n_hidden_features,
            activation_name=activation_name,
            a=a,
            nodes_sim=nodes_sim,
            bias=bias,
            dropout=dropout,
            direct_link=direct_link,
            n_clusters=n_clusters,
            cluster_encode=cluster_encode,
            type_clust=type_clust,
            type_scaling=type_scaling,
            lambda1=lambda1,
            lambda2=lambda2,
            seed=seed,
            backend=backend,
        )

        self.type_fit = "classification"

    def loglik(self, X, Y, **kwargs):
        """Log-likelihood for training data (X, Y).

        Args:

            X: {array-like}, shape = [n_samples, n_features]
                Training vectors, where n_samples is the number
                of samples and n_features is the number of features.

            Y: array-like, shape = [n_samples]
                One-hot encode target values.

            **kwargs: additional parameters to be passed to
                    self.cook_training_set or self.obj.fit

        Returns:

        """

        def loglik_grad_hess(Y, X, B, XB, hessian=True, **kwargs):
            # nobs, n_classes
            n, K = Y.shape

            # total number of covariates
            p = X.shape[1]

            # initial number of covariates
            init_p = p - self.n_hidden_features

            max_double = 709.0
            XB[XB > max_double] = max_double
            exp_XB = np.exp(XB)
            probs = exp_XB / exp_XB.sum(axis=1)[:, None]

            # gradient -----
            # (Y - p) -> (n, K)
            # X -> (n, p)
            # (K, n) %*% (n, p) -> (K, p)
            if hessian is False:
                grad = (
                    -mo.safe_sparse_dot(
                        a=(Y - probs).T, b=X, backend=self.backend
                    )
                    / n
                )
                grad += self.lambda1 * B[0:init_p, :].sum(axis=0)[:, None]
                grad += self.lambda2 * B[init_p:p, :].sum(axis=0)[:, None]

                return grad.flatten()

            # hessian -----
            if hessian is True:
                Kp = K * p
                hess = np.zeros((Kp, Kp), float)
                for k1 in range(K):
                    x_index = range(k1 * p, (k1 + 1) * p)
                    for k2 in range(k1, K):
                        y_index = range(k2 * p, (k2 + 1) * p)
                        H_sub = (
                            -mo.safe_sparse_dot(
                                a=X.T,
                                b=(probs[:, k1] * probs[:, k2])[:, None] * X,
                                backend=self.backend,
                            )
                            / n
                        )  # do not store
                        hess[np.ix_(x_index, y_index)] = hess[
                            np.ix_(y_index, x_index)
                        ] = H_sub

                return hess + (self.lambda1 + self.lambda2) * np.identity(Kp)

        # total number of covariates
        p = X.shape[1]

        # initial number of covariates
        init_p = p - self.n_hidden_features

        # log-likelihood (1st return)
        def loglik_func(x):
            # (p, K)
            B = x.reshape(Y.shape[1], p).T

            # (n, K)
            XB = mo.safe_sparse_dot(X, B, backend=self.backend)

            res = -(np.sum(Y * XB, axis=1) - logsumexp(XB)).mean()

            res += (
                0.5
                * self.lambda1
                * mo.squared_norm(B[0:init_p, :], backend=self.backend)
            )
            res += (
                0.5
                * self.lambda2
                * mo.squared_norm(B[init_p:p, :], backend=self.backend)
            )

            return res

        # gradient of log-likelihood
        def grad_func(x):
            # (p, K)
            B = x.reshape(Y.shape[1], p).T

            return loglik_grad_hess(
                Y=Y,
                X=X,
                B=B,
                XB=mo.safe_sparse_dot(X, B, backend=self.backend),
                hessian=False,
                **kwargs
            )

        # hessian of log-likelihood
        def hessian_func(x):
            # (p, K)
            B = x.reshape(Y.shape[1], p).T

            return loglik_grad_hess(
                Y=Y,
                X=X,
                B=B,
                XB=mo.safe_sparse_dot(X, B, backend=self.backend),
                hessian=True,
                **kwargs
            )

        return loglik_func, grad_func, hessian_func

    # newton-cg
    # L-BFGS-B
    def fit(self, X, y, solver="L-BFGS-B", **kwargs):
        """Fit Ridge model to training data (X, y).

        for beta: regression coeffs (beta11, ..., beta1p, ..., betaK1, ..., betaKp)
        for K classes and p covariates.

        Args:

            X: {array-like}, shape = [n_samples, n_features]
                Training vectors, where n_samples is the number
                of samples and n_features is the number of features.

            y: array-like, shape = [n_samples]
                Target values.

            **kwargs: additional parameters to be passed to
                    self.cook_training_set or self.obj.fit

        Returns:

            self: object

        """

        assert mx.is_factor(y), "y must contain only integers"

        self.n_classes_ = len(np.unique(y)) # for compatibility with sklearn 

        output_y, scaled_Z = self.cook_training_set(y=y, X=X, **kwargs)

        self.n_classes = len(np.unique(y))

        Y = mo.one_hot_encode2(output_y, self.n_classes)

        # optimize for beta, minimize self.loglik (maximize loglik) -----
        loglik_func, grad_func, hessian_func = self.loglik(X=scaled_Z, Y=Y)

        if solver == "L-BFGS-B":
            self.beta_ = minimize(
                fun=loglik_func,
                x0=np.zeros(scaled_Z.shape[1] * self.n_classes),
                jac=grad_func,
                method=solver,
            ).x

        if solver in ("Newton-CG", "trust-ncg"):
            self.beta_ = minimize(
                fun=loglik_func,
                x0=np.zeros(scaled_Z.shape[1] * self.n_classes),
                jac=grad_func,
                hess=hessian_func,
                method=solver,
            ).x
        self.classes_ = np.unique(y)
        return self

    def predict(self, X, **kwargs):
        """Predict test data X.

        Args:

            X: {array-like}, shape = [n_samples, n_features]
                Training vectors, where n_samples is the number
                of samples and n_features is the number of features.

            **kwargs: additional parameters to be passed to
                    self.cook_test_set

        Returns:

            model predictions: {array-like}
        """

        return np.argmax(self.predict_proba(X, **kwargs), axis=1)

    def predict_proba(self, X, **kwargs):
        """Predict probabilities for test data X.

        Args:

            X: {array-like}, shape = [n_samples, n_features]
                Training vectors, where n_samples is the number
                of samples and n_features is the number of features.

            **kwargs: additional parameters to be passed to
                    self.cook_test_set

        Returns:

            probability estimates for test data: {array-like}

        """
        if len(X.shape) == 1:
            n_features = X.shape[0]
            new_X = mo.rbind(
                X.reshape(1, n_features),
                np.ones(n_features).reshape(1, n_features),
            )

            Z = self.cook_test_set(new_X, **kwargs)

        else:
            Z = self.cook_test_set(X, **kwargs)

        ZB = mo.safe_sparse_dot(
            a=Z,
            b=self.beta_.reshape(
                self.n_classes,
                X.shape[1] + self.n_hidden_features + self.n_clusters,
            ).T,
            backend=self.backend,
        )

        exp_ZB = np.exp(ZB)

        return exp_ZB / exp_ZB.sum(axis=1)[:, None]


from .lmfuncs import beta_hat, inv_penalized_cov
from .matrixops import (
    cbind,
    delete_last_columns,
    rbind,
    convert_df_to_numeric,
    crossprod,
    tcrossprod,
    to_np_array,
)
from .misc import merge_two_dicts, is_factor, tuple_map
from .model_selection import TimeSeriesSplit
from .progress_bar import Progbar
from .psdcheck import isPD, nearestPD
from .timeseries import (
    compute_output_dates,
    create_lags,
    create_train_inputs,
    reformat_response,
)


__all__ = [
    "beta_hat",
    "inv_penalized_cov",
    "cbind",
    "delete_last_columns",
    "rbind",
    "convert_df_to_numeric",
    "crossprod",
    "tcrossprod",
    "to_np_array",
    "merge_two_dicts",
    "is_factor",
    "isPD",
    "nearestPD",
    "compute_output_dates",
    "create_lags",
    "create_train_inputs",
    "reformat_response",
    "tuple_map",
    "TimeSeriesSplit",
    "Progbar",    
]
